"""Scalar plane wave module.

.. autosummary::
   :toctree:

   to_scw
   to_ssw
   translate
   permute_xyz

"""

import numpy as np
import treams.special as sc
import scipy.special as ss


def translate(kx, ky, kz, x, y, z, *args, **kwargs):
    r"""translate(kx, ky, kz, x, y, z)
    
    Translation coefficients for scalar plane waves.

    The translation coefficient is the phase factor
    :math:`\mathrm e^{\mathrm i \mathbf k \mathbf r}`.

    Args:
        kx (float or complex or array_like): X-component of the wave vector.
        ky (float or complex or array_like): Y-component of the wave vector.
        kz (float or complex or array_like): Z-component of the wave vector.
        x (float or array_like): X-coordinate. Has units of 1/kx.
        y (float or array_like): Y-coordinate. Has units of 1/ky.
        z (float or array_like): Z-coordinate. Has units of 1/kz.

    Returns:
        complex or array_like
    """
    return np.exp(1j * (kx * x + ky * y + kz * z), *args, **kwargs)


def _to_ssw(l, m, kx, ky, kz, *args, **kwargs):
    """Coefficient for the expansion of a scalar plane wave in scalar spherical waves"""
    phi = np.arctan2(ky, kx)
    k = np.sqrt(kx * kx + ky * ky + kz * kz)
    pref = (
        2 * np.sqrt(np.pi * (2 * l + 1))
        * np.sqrt(ss.gamma(l - m + 1) / ss.gamma(l + m + 1))
        * np.power(1j, l)
        * np.exp(-1j * m * phi)
    )
    return pref * sc.lpmv(m, l, kz / k, *args, **kwargs)

_to_ssw = np.vectorize(_to_ssw)

def to_ssw(l, m, kx, ky, kz, *args, **kwargs):
    """to_ssw(l, m, kx, ky, kz)
    
    Coefficients for the expansion of scalar plane waves in scalar spherical waves.

    Return the coefficient for the basis change from scalar plane waves to scalar spherical waves.
    For multiple positions, only the diagonal values (corresponding to identical positions) are
    returned.

    Args:
        l (int or array_like): Degree of scalar spherical waves.
        m (int or array_like): Order of scalar spherical waves.
        kx (float or array_like): X-component of the wave vector of scalar plane waves.
        ky (float or array_like): Y-component of the wave vector of scalar plane waves. Has units of kx.
        kz (float or array_like): Z-component of the wave vector of scalar plane waves. has units of kx.
 
    Returns:
        complex or array_like
    """
    return _to_ssw(l, m, kx, ky, kz, *args, **kwargs)

    
def _to_scw(kzcw, m, kx, ky, kzpw, *args, **kwargs):
    """Coefficient for the expansion of a scalar plane wave in scalar cylindricrical waves"""  
    krho = np.sqrt(kx * kx + ky * ky)
    if np.abs(kzcw - kzpw) <= 1e-12:
        if m == 0:
            return np.power(1, 0, *args, **kwargs)
        if krho == 0:
            return np.power(1j, m, *args, **kwargs)
        return np.power((1j * kx + ky) / krho, m, *args, **kwargs)
    elif np.abs(kzcw - kzpw) > 1e-12:
        return 0.0j + sc.lpmv(0, 1, 0, *args, **kwargs)
    

_to_scw = np.vectorize(_to_scw)    

def to_scw(kzcw, m, kx, ky, kzpw, *args, **kwargs):
    """to_scw(qz, m, kx, ky, kz)

    Coefficients for the expansion of scalar plane waves in scalar cylindrical waves.

    Return the coefficient for the basis change from scalar plane waves to scalar cylindrical waves.
    For multiple positions, only the diagonal values (corresponding to identical positions) are returned.

    Args:
        qz (float or array_like): Z-component of the wave vector of scalar cylindrical waves.
        m (int or array_like): Order of scalar cylindrical waves.
        kx (float or array_like): X-component of the wave vector of scalar plane waves. Has units of qz.
        ky (float or array_like): Y-component of the wave vector of scalar plane waves. Has units of qz.
        kz (float or array_like): Z-component of the wave vector of scalar plane waves. Has units of qz.

    Returns:
        complex or array_like
    """
    return _to_scw(kzcw, m, kx, ky, kzpw, *args, **kwargs)


def _xyz_to_zxy(kx, ky, kz, *args, **kwargs):
    return np.power(1, 0, *args, **kwargs)

_xyz_to_zxy = np.vectorize(_xyz_to_zxy)  

def _xyz_to_yzx(kx, ky, kz, *args, **kwargs):
    return np.power(1, 0, *args, **kwargs)

_xyz_to_yzx = np.vectorize(_xyz_to_yzx)

def permute_xyz(kx, ky, kz, inverse=False, *args, **kwargs):
    """permute_xyz(kx, ky, kz, inverse=False)
    
    Change the coordinate system of the plane-wave basis.

    A plane wave in the coordinate system :math:`(x, y, z)` with the primary direction of
    propagation along the z-axis is described in the system :math:`(x', y', z') = (y, z, x)`. 
    The inverse transformation is also possible.

    The function is essentially diagonal relative to the wavenumber, because we always describe
    the input and output modes in the unprimed coordinate system.

    Args:
        kx (float or array_like): X-component of the wave vector.
        ky (float or array_like): Y-component of the wave vector.
        kz (float or complex or array_like): Z-component of the wave vector.
        inverse (bool, optional): Defaults to False. If True, use the inverse transformation.

    Returns:
        complex or array_like
    """
    if inverse:
        return _xyz_to_zxy(kx, ky, kz, *args, **kwargs)
    return _xyz_to_yzx(kx, ky, kz, *args, **kwargs)


import copy

import numpy as np

from treams import util
from acoustotreams._coreacoustics import AcousticsArray
from acoustotreams._coreacoustics import ScalarPlaneWaveBasisByComp as SPWBC
from acoustotreams._materialacoustics import AcousticMaterial
from acoustotreams._operatorsacoustics import translate
from acoustotreams._tmatrixacoustics import AcousticTMatrixC
from acoustotreams.coeffs import fresnel_acoustics


class AcousticSMatrix(AcousticsArray):
    """Acoustic S-matrix for a scalar plane wave."""

    def _check(self):
        super()._check()
        shape = np.shape(self)
        if len(shape) != 2 or shape[0] != shape[1]:
            raise util.AnnotationError(f"invalid shape: '{shape}'")
        if not isinstance(self.k0, (int, float, np.floating, np.integer)):
            raise util.AnnotationError("invalid k0")
        material = (None, None) if self.material is None else self.material
        if isinstance(material, tuple):
            self.material = tuple(AcousticMaterial() if m is None else m for m in material)
        if self.basis is None:
            raise util.AnnotationError("basis not set")
        if not isinstance(self.basis, SPWBC):
            self.basis = self.basis.bycomp(self.k0, self.material)


class OperatorAttributeSMatrices:
    def __init__(self, name):
        self._name = name
        self._obj = self._objtype = None

    def __get__(self, obj, objtype=None):
        self._obj = obj
        self._objtype = type(obj) if objtype is None else objtype
        return self

    def __call__(self, *args, **kwargs):
        res = [
            [getattr(ii, self._name)(*args, **kwargs) for ii in i] for i in self._obj
        ]
        return self._objtype(res)

    def apply_left(self, *args, **kwargs):
        res = [
            [getattr(ii, self._name).apply_left(*args, **kwargs) for ii in i]
            for i in self._obj
        ]
        return self._objtype(res)

    def apply_right(self, *args, **kwargs):
        res = [
            [getattr(ii, self._name).apply_right(*args, **kwargs) for ii in i]
            for i in self._obj
        ]
        return self._objtype(res)


class AcousticSMatrices:
    r"""Collection of four acoustic S-matrices with a scalar plane-wave basis.

    The S-matrix describes the scattering of incoming into outgoing modes using a 
    plane-wave basis, given by functions :func:`acoustotreams.speical.spw_Psi`. The primary
    direction of propagation is parallel or anti-parallel to the z-axis. The scattering
    structure itself is infinitely elongated in the x- and y-directions. The S-matrix 
    contains four blocks :math:`S_{\uparrow \uparrow}`, :math:`S_{\uparrow \downarrow}`, 
    :math:`S_{\downarrow \uparrow}`, and :math:`S_{\downarrow \downarrow}`:

    .. math::

        S = \begin{pmatrix}
            S_{\uparrow \uparrow} & S_{\uparrow \downarrow} \\
            S_{\downarrow \uparrow} & S_{\downarrow \downarrow}
        \end{pmatrix}\,.

    These matrices describe the transmission of plane waves propagating into positive
    z-direction, reflection of plane waves into the positive z-direction, reflection of
    plane waves into negative z-direction, and the transmission of plane waves
    propagating in negative z-direction, respectively. Each of these four matrices
    also maps different diffraction orders to one another.

    The wavenumber :attr:`k0` and, if not the air, the material :attr:`material` 
    are required.

    Args:
        smats (AcousticSMatrix): Acoustic S-matrices.
        k0 (float): Angular wavenumber in the air (rad/m).
        basis (ScalarPlaneWaveBasisByComp): Basis for the S-matrices.
        material (tuple or AcousticMaterial, optional): Material definition. If a tuple of length
            two is specified, it refers to the materials above and below the S-matrix.
    """
    permute = OperatorAttributeSMatrices("permute")
    translate = OperatorAttributeSMatrices("translate")
    rotate = OperatorAttributeSMatrices("rotate")

    def __init__(self, smats, **kwargs):
        """Initialization."""
        if len(smats) != 2 or len(smats[0]) != 2 or len(smats[1]) != 2:
            raise ValueError("invalid shape of S-matrices")
        if "material" in kwargs:
            material = kwargs["material"]
            del kwargs["material"]
        elif hasattr(smats[0][0], "material"):
            material = smats[0][0].material
        elif hasattr(smats[1][1], "material"):
            material = smats[1][1].material
            if isinstance(material, tuple):
                material = material[::-1]
        elif hasattr(smats[0][1], "material"):
            material = smats[0][1].material
        elif hasattr(smats[1][0], "material"):
            material = smats[1][0].material
        else:
            material = AcousticMaterial()
        if isinstance(material, tuple):
            ma, mb = material
        else:
            ma = mb = AcousticMaterial(material)
        material = [[(ma, mb), (ma, ma)], [(mb, mb), (mb, ma)]]
        modetype = [[("up", "up"), ("up", "down")], [("down", "up"), ("down", "down")]]
        self._sms = [
            [
                AcousticSMatrix(s, material=m, modetype=t, **kwargs)
                for s, m, t in zip(ar, mr, tr)
            ]
            for ar, mr, tr in zip(smats, material, modetype)
        ]
        self.material = self[0, 0].material
        self.basis = self[0, 0].basis
        self.k0 = self[0, 0].k0
        for i in (self[1, 0], self[0, 1], self[1, 1]):
            i.k0 = self.k0
            i.basis = self.basis

    def __eq__(self, other):
        return all(
            (np.abs(self[i, j] - other[i][j]) < 1e-14).all()           
            for i in range(2)
            for j in range(2)
        )

    def __getitem__(self, key):
        keys = {0: 0, "up": 0, 1: 1, "down": 1}
        if key in keys:
            return self._sms[keys[key]]
        if not isinstance(key, tuple) or len(key) not in (1, 2):
            raise KeyError(f"invalid key: '{key}'")
        if len(key) == 1:
            return self[key[0]]
        key = tuple(keys[k] for k in key)
        return self._sms[key[0]][key[1]]

    def __len__(self):
        return 2

    def __iter__(self):
        return iter(self._sms)
    
    @classmethod
    def interface(cls, basis, k0, materials):
        """Planar interface between two media.

        Args:
            basis (ScalarPlaneWaveBasisByComp): Basis definitions.
            k0 (float): Angular wavenumber in the air (rad/m).
            materials (Sequence[AcousticMaterial]): Material definitions.

        Returns:
            AcousticSMatrix
        """
        materials = tuple(map(AcousticMaterial, materials))
        qs = np.zeros((2, 2, len(basis), len(basis)), complex)
        for i, (kx, ky) in enumerate(basis):
            kzs = np.array([m.kzs(k0, kx, ky) for m in materials])
            vals = fresnel_acoustics(kzs, [m.rho for m in materials])
            qs[:, :, i, i] = vals[:, :, 0, 0]
        return cls(qs, k0=k0, basis=basis, material=materials[::-1])

    @classmethod
    def slab(cls, thickness, basis, k0, materials):
        """Slab of material.

        A slab of material, defined by a thickness and three materials. Consecutive
        slabs of material can be defined by `n` thicknesses and `(n + 2)` material
        parameters.

        Args:
            k0 (float): Angular wavenumber in air (rad/m).
            basis (ScalarPlaneWaveBasisByComp): Basis definition.
            tickness (Sequence[Float]): Thickness of the slab or the thicknesses of all
                slabs in order from negative to positive z.
            materials (Sequence[AcousticMaterial]): Material definitions from negative to
                positive z.

        Returns:
            AcousticSMatrix
        """
        try:
            iter(thickness)
        except TypeError:
            thickness = [thickness]
        res = cls.interface(basis, k0, materials[:2])
        for d, ma, mb in zip(thickness, materials[1:-1], materials[2:]):
            if np.ndim(d) == 0:
                d = [0, 0, d]
            x = cls.propagation(d, basis, k0, ma)
            res = res.add(x)
            res = res.add(cls.interface(basis, k0, (ma, mb)))
        return res

    @classmethod
    def stack(cls, items):
        """Stack of S-matrices.

        Couple multiple S-matrices in the order given. Before
        coupling it can be checked for matching materials and modes.

        Args:
            items (Sequence[AcousticSMatrix]): An array of S-matrices in their intended order
                from negative to positive z.

        Returns:
            AcousticSMatrix
        """
        acc = items[0]
        for item in items[1:]:
            acc = acc.add(item)
        return acc

    @classmethod
    def propagation(cls, r, basis, k0, material=AcousticMaterial()):
        """S-matrix for the propagation along a distance.

        This S-matrix translates the reference origin along `r`.

        Args:
            r (float, (3,)-array): Translation vector.
            basis (ScalarPlaneWaveBasis): Basis definition.
            k0 (float): Angular wavenumber in the air. Has units of 1/r.
            material (AcousticMaterial, optional): Material definition.

        Returns:
            AcousticSMatrix
        """
        sup = translate(r, basis=basis, k0=k0, material=material, modetype="up")
        sdown = translate(
            np.negative(r), basis=basis, k0=k0, material=material, modetype="down"
        )
        zero = np.zeros_like(sup)
        material = AcousticMaterial(material)
        return cls([[sup, zero], [zero, sdown]], basis=basis, k0=k0, material=material)

    @classmethod
    def from_array(cls, tm, basis):
        """S-matrix from an array of (cylindrical) T-matrices.

        Create a S-matrix for a two-dimensional array of scatterers described by the
        T-Matrix or an one-dimensional array of scatterers described by the cylindrical
        T-matrix.

        Args:
            tm (AcousticTMatrix or AcousticTMatrixC): (Cylindrical) T-matrix to place on array.
            basis (ScalarPlaneWaveBasisByComp): Basis definition.
            eta (float or complex, optional): Splitting parameter in the lattice sum.

        Returns:
            SMatrix
        """
        if isinstance(tm, AcousticTMatrixC):
            basis = basis.permute(-1)
        pu, pd = (tm.expandlattice(basis=basis, modetype=i) for i in ("up", "down"))
        au, ad = (tm.expand.eval_inv(basis, i) for i in ("up", "down"))
        eye = np.eye(len(basis))
        res = cls([[eye + pu @ au, pu @ ad], [pd @ au, eye + pd @ ad]])
        if isinstance(tm, AcousticTMatrixC):
            res = res.permute()
        return res

    def add(self, sm):
        """Couple another S-matrix on top of the current one.

        See also :func:`AcousticSMatrix.stack` for a function that does not change the
        current S-matrix but creates a new one.

        Args:
            sm (AcousticSMatrix): Acoustic S-matrix to add.

        Returns:
            AcousticSMatrix
        """
        dim = len(self.basis)
        snew = [[None, None], [None, None]]
        s_tmp = np.linalg.solve(np.eye(dim) - self[0, 1] @ sm[1][0], self[0, 0])
        snew[0][0] = sm[0][0] @ s_tmp
        snew[1][0] = self[1, 0] + self[1, 1] @ sm[1][0] @ s_tmp
        s_tmp = np.linalg.solve(np.eye(dim) - sm[1][0] @ self[0, 1], sm[1][1])
        snew[1][1] = self[1, 1] @ s_tmp
        snew[0][1] = sm[0][1] + sm[0][0] @ self[0, 1] @ s_tmp
        return AcousticSMatrices(snew)

    def double(self, n=1):
        """Double an acoustic S-matrix.

        By default, this function doubles the S-matrix, but it can also create a
        :math:`2^n`-fold repetition of itself:

        Args:
            n (int, optional): Number of times to double itself. Defaults to 1.

        Returns:
            AcousticSMatrix
        """
        res = self
        for _ in range(n):
            res = res.add(self)
        return res

    def illuminate(self, illu, illu2=None, /, *, smat=None):
        """Field coefficients above and below the S-matrix.

        Given an insonification defined by the coefficients of each incoming mode,
        compute the coefficients of the outgoing modes above and below the S-matrix.
        If a second S-matrix is given, the field expansions in between are also
        computed.

        Args:
            illu (complex or array_like): Insonification. If `modetype` is specified, the direction
                is chosen accordingly.
            illu2 (complex or array_like, optional): Second insonification. If used, the first
                argument is taken to be coming from below and this one to be coming from
                above.
            smat (AcousticSMatrix, optional): Second S-matrix for computing the
                field expansion between two S-matrices.

        Returns:
            tuple
        """
        modetype = getattr(illu, "modetype", "up")
        if isinstance(modetype, tuple):
            modetype = modetype[max(-2, -len(tuple))]
        illu2 = np.zeros(np.shape(illu)[-2:]) if illu2 is None else illu2
        if modetype == "down":
            illu, illu2 = illu2, illu
        if smat is None:
            field_up = self[0, 0] @ illu + self[0, 1] @ illu2
            field_down = self[1, 0] @ illu + self[1, 1] @ illu2
            return field_up, field_down
        stmp = np.eye(len(self.basis)) - self[0, 1] @ smat[1, 0]
        field_in_up = np.linalg.solve(
            stmp, self[0, 0] @ illu + self[0, 1] @ smat[1, 1] @ illu2
        )
        field_in_down = smat[1, 0] @ field_in_up + smat[1, 1] @ illu2
        field_up = smat[0, 1] @ illu2 + smat[0, 0] @ field_in_up
        field_down = self[1, 0] @ illu + self[1, 1] @ field_in_down
        return field_up, field_down, field_in_up, field_in_down

    def tr(self, illu):
        """Transmittance and reflectance.

        Calculate the transmittance and reflectance for one S-matrix with the given
        insonification and propagation direction.

        Args:
            illu (complex or array_like): Expansion coefficients of the incoming wave.

        Returns:
            tuple
        """
        modetype = getattr(illu, "modetype", "up")
        if isinstance(modetype, tuple):
            modetype = modetype[max(-2, -len(tuple))]
        trans, refl = self.illuminate(illu)
        material = self.material
        if not isinstance(material, tuple):
            material = material, material
        paz = [poynting_avg_z(self.basis, self.k0, m) for m in material]
        if modetype == "down":
            trans, refl = refl, trans
            paz.reverse()
        illu = np.asarray(illu)
        s_t = np.real(trans.conjugate().T @ paz[0] @ trans)             
        s_r = np.real(refl.conjugate().T @ paz[1] @ refl)
        s_i = np.real(np.conjugate(illu).T @ paz[1] @ illu)
        s_ir = np.real(
            refl.conjugate().T @ paz[1] @ illu
            - np.conjugate(illu).T @ paz[1] @ refl
        )
        return s_t / (s_i + s_ir), s_r / (s_i + s_ir)

    def periodic(self):
        r"""Periodic repetition of an S-matrix.

        Transform the S-matrix to an infinite periodic arrangement of itself defined
        by

        .. math::

            \begin{pmatrix}
                S_{\uparrow \uparrow} & S_{\uparrow \downarrow} \\
                -S_{\downarrow \downarrow}^{-1}
                S_{\downarrow \uparrow} S_{\uparrow \uparrow} &
                S_{\downarrow \downarrow}^{-1} (\mathbb{1} - S_{\downarrow \uparrow}
                S_{\uparrow \downarrow})
            \end{pmatrix}\,.

        Returns:
            complex or array_like

        """
        dim = len(self.basis)
        res = np.empty((2 * dim, 2 * dim), dtype=complex)
        res[0:dim, 0:dim] = self[0, 0]
        res[0:dim, dim:] = self[0, 1]
        res[dim:, 0:dim] = -np.linalg.solve(self[1, 1], self[1, 0] @ self[0, 0])
        res[dim:, dim:] = np.linalg.solve(
            self[1, 1], np.eye(dim) - self[1, 0] @ self[0, 1]
        )
        return res

    def bands_kz(self, az):
        r"""Band structure calculation.

        Calculate the band structure for a given S-matrix, assuming it is periodically
        repeated along the z-axis. The function returns the z-components of the wave
        vector :math:`k_z` and the corresponding eigenvectors :math:`v` of

        .. math::

            \begin{pmatrix}
                S_{\uparrow \uparrow} & S_{\uparrow \downarrow} \\
                -S_{\downarrow \downarrow}^{-1} S_{\downarrow \uparrow}
                    S_{\uparrow \uparrow} &
                S_{\downarrow \downarrow}^{-1} (\mathbb{1} - S_{\downarrow \uparrow}
                    S_{\uparrow \downarrow})
            \end{pmatrix}
            \boldsymbol v
            =
            \mathrm{e}^{\mathrm{i}k_z a_z}
            \boldsymbol v\,.

        Args:
            az (float): Lattice pitch along the z-direction (m).

        Returns:
            tuple

        """
        w, v = np.linalg.eig(self.periodic())
        return -1j * np.log(w) / az, v

    def __repr__(self):
        return f"""{type(self).__name__}(...,
    basis={self.basis},
    k0={self.k0},
    material={self.material},
)"""

def poynting_avg_z(basis, k0, material=AcousticMaterial()):
    r"""Time-averaged z-component of the Poynting vector.

    Return the time-averaged Poynting vector's z-component

    .. math::

        \langle S_z \rangle
        = \frac{1}{2} \Re (p v_z^\ast) \boldsymbol{\hat{z}}

    on one side of the S-matrix with the given coefficients.

    Returns:
        tuple
    """
    kx, ky, kzs = basis.kvecs(k0, material)
    selection = (kx[:, None] == kx) & (ky[:, None] == ky)
    gamma = kzs / (material.ks(k0) * material.impedance)
    return 0.5 * np.real(selection * gamma)


"""ACOUSTOTREAMS: A Python package for acoustic wave scattering based on the T-matrix method.

.. currentmodule:: acoustotreams

Classes
=======

The top-level classes and functions allow a high-level access to the functionality.

Basis sets
----------

.. autosummary::
   :toctree: generated/

   ScalarCylindricalWaveBasis
   ScalarPlaneWaveBasisByUnitVector
   ScalarPlaneWaveBasisByComp
   ScalarSphericalWaveBasis

Matrices and Arrays
-------------------

.. autosummary::
   :toctree: generated/

   AcousticsArray
   AcousticSMatrix
   AcousticSMatrices
   AcousticTMatrix
   AcousticTMatrixC

Other
-----

.. autosummary::
   :toctree: generated/

   AcousticMaterial

Functions
=========

Operators
---------

.. autosummary::
   :toctree: generated/

   pfield
   vfield
   pamplitudeff
   expand
   expandlattice
   permute
   rotate
   translate

Scalar wave functions
---------------------

.. autosummary::
   :toctree: generated/

   cylindrical_wave_scalar
   plane_wave_scalar
   plane_wave_angle_scalar
   spherical_wave_scalar

Functions imported from SciPy
-----------------------------

+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.hankel1`\(v, z[, out])            | Hankel function of the  |
|                                                            | first kind.             |
+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.hankel2`\(v, z[, out])            | Hankel function of the  |
|                                                            | second kind.            |
+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.jv`\(v, z[, out])                 | Bessel function of the  |
|                                                            | first kind of real      |
|                                                            | order and complex       |
|                                                            | argument.               |
+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.yv`\(v, z[, out])                 | Bessel function of the  |
|                                                            | second kind of real     |
|                                                            | order and complex       |
|                                                            | argument.               |
+------------------------------------------------------------+-------------------------+
| | :py:func:`spherical_jn <scipy.special.spherical_jn>`\(n, | Spherical Bessel        |
|   z[, derivative])                                         | function of the first   |
|                                                            | kind or its derivative. |
+------------------------------------------------------------+-------------------------+
| | :py:func:`spherical_yn <scipy.special.spherical_yn>`\(n, | Spherical Bessel        |
|   z[, derivative])                                         | function of the second  |
|                                                            | kind or its derivative. |
+------------------------------------------------------------+-------------------------+

Functions imported from treams.special, treams.misc, and treams.lattice
-----------------------------------------------------------------------

+------------------------------------------------------------+-----------------------------+
| :py:data:`~treams.special.spherical_jn_d`\(n, z)           | Derivative of the spherical | 
|                                                            | Bessel function of the      |
|                                                            | first kind.                 |   
+------------------------------------------------------------+-----------------------------+
| :py:data:`~treams.special.spherical_yn_d`\(n, z)           | Derivative of the spherical | 
|                                                            | Bessel function of the      |
|                                                            | second kind.                |   
+------------------------------------------------------------+-----------------------------+

"""

_version__ = "0.2.49"

from treams.misc import(  # noqa: F401,
    wave_vec_z,
    firstbrillouin1d,
    firstbrillouin2d,
    firstbrillouin3d,
)

from treams._lattice import *   # noqa: F401

from acoustotreams._materialacoustics import AcousticMaterial  # noqa: F401

from acoustotreams._smatrixacoustics import (  # noqa: F401
    AcousticSMatrices,
    AcousticSMatrix,
    poynting_avg_z,
)

import acoustotreams.scw  # noqa: F401

from acoustotreams._coreacoustics import (  # noqa: F401
    ScalarCylindricalWaveBasis,
    AcousticsArray,
    ScalarPlaneWaveBasisByComp,
    ScalarPlaneWaveBasisByUnitVector,
    ScalarSphericalWaveBasis,
)

from acoustotreams._tmatrixacoustics import (  # noqa: F401
    AcousticTMatrix,
    AcousticTMatrixC,
    cylindrical_wave_scalar,
    plane_wave_scalar,
    plane_wave_angle_scalar,
    spherical_wave_scalar,
)

import acoustotreams.coeffs  # noqa: F401 

import acoustotreams.spw  # noqa: F401

from acoustotreams._operatorsacoustics import (  # noqa: F401
    PField,
    VField,
    PAmplitudeFF,
    Expand,
    ExpandLattice,
    Permute,
    Rotate,
    Translate,
    vfield,
    pfield,
    pamplitudeff,
    expand,
    expandlattice,
    permute,
    rotate,
    translate,
)

import acoustotreams.ssw # noqa: F401



"""Scattering coefficients for high-symmetry cases.

Calculate the scattering coefficients for cases where they can be obtained analytically.
Thes include a (multilayered) sphere using spherical waves (Mie coefficients), 
a cylinder using cylindrical waves, and an infinitely extended planar
interface (Fresnel coefficients).

.. autosummary::
   :toctree:

   mie_acoustics
   mie_acoustics_cyl
   fresnel_acoustics

"""

import treams.special
import numpy as np
from acoustotreams._materialacoustics import AcousticMaterial

def _mie_acoustics_iter(tm, l, x, mat_sphere, mat_env):
    """Solve a matrix for the interface conditions at a spherical interface.

    Six types of interfaces are supported: fluid-fluid, fluid-solid, solid-fluid,
    solid-solid, soft-fluid, and hard-fluid.

    The fluid medium has :math:`(\rho, c, c_t = 0)`, the solid medium has
    :math:`(\rho, c, c_t)`, the soft medium has :math:`(\rho = 0, c = 0, c_t = 0)`,
    and the hard medium has :math:`(\rho = \infty, c = 0, c_t = 0)`.

    Note:
        The order of T-matrix entries is LL, NL, LN, and NN.

    Args:
        tm (array_like of float or complex): T-matrix of the previous layer.
        l (int): Degree of the coefficient.
        x (float): Size parameters in the air.
        mat_sphere (array_like of float or complex): Material of the sphere :math:`(\rho, c, c_t)`.
        mat_env (array_like of float or complex): Material of the medium :math:`(\rho, c, c_t)`.

    Returns:
        complex 4-array
    """ 
    x_env = x * AcousticMaterial().c / mat_env[1]

    j = treams.special.spherical_jn(l, x_env, derivative=False)
    h = treams.special.spherical_hankel1(l, x_env)
    if (np.abs(mat_sphere[2]) == 0 
        and np.abs(mat_sphere[1]) == 0 
        and np.abs(mat_sphere[0]) == 0
        and np.abs(mat_env[2]) == 0):
        return np.array([-j / h, 0, 0, 0])

    j_d = treams.special.spherical_jn(l, x_env, derivative=True)
    h_d = treams.special.spherical_hankel1_d(l, x_env)
    if (np.abs(mat_sphere[2]) == 0 
        and np.abs(mat_sphere[1]) == 0 
        and np.abs(mat_sphere[0]) == np.inf
        and np.abs(mat_env[2]) == 0):
        return np.array([-j_d / h_d, 0, 0, 0])
    
    x_sphere = x * AcousticMaterial().c / mat_sphere[1]
    j1 = treams.special.spherical_jn(l, x_sphere, derivative=False)
    j1_d = treams.special.spherical_jn(l, x_sphere, derivative=True)
    h1 = treams.special.spherical_hankel1(l, x_sphere)
    h1_d = treams.special.spherical_hankel1_d(l, x_sphere)

    psi = np.sqrt(l * (l + 1))
    d12 = h
    d22 = x_env * h_d
    d32 = x_env * h_d - h
    d14 = j1
    f14 = h1
    d24 = x_sphere * j1_d
    f24 = x_sphere * h1_d
    d1L = j
    d2L = x_env * j_d
    d3L = x_env * j_d - j
    
    res = np.zeros(4, complex)

    if np.abs(mat_env[2]) > 0:
        x_env_t = x * AcousticMaterial().c / mat_env[2]
        d42 = (psi**2 - 0.5 * x_env_t**2) * h - 2 * x_env * h_d
        d4L = (psi**2 - 0.5 * x_env_t**2) * j - 2 * x_env * j_d
        if l > 0:
            jt = treams.special.spherical_jn(l, x_env_t, derivative=False)
            jt_d = treams.special.spherical_jn(l, x_env_t, derivative=True)
            ht = treams.special.spherical_hankel1(l, x_env_t)
            ht_d = treams.special.spherical_hankel1_d(l, x_env_t)
            d21 = psi**2 * ht 
            d31 = (psi**2 - 0.5 * x_env_t**2 - 1) * ht - x_env_t * ht_d
            d41 = psi**2 * (x_env_t * ht_d - ht)
            d2N = psi**2 * jt
            d3N = (psi**2 - 0.5 * x_env_t**2 - 1) * jt - x_env_t * jt_d
            d4N = psi**2 * (x_env_t * jt_d - jt) 

    if np.abs(mat_sphere[2]) > 0:
        x_sphere_t = x * AcousticMaterial().c / mat_sphere[2]
        if l > 0:
            j1t = treams.special.spherical_jn(l, x_sphere_t, derivative=False)
            j1t_d = treams.special.spherical_jn(l, x_sphere_t, derivative=True)
            h1t = treams.special.spherical_hankel1(l, x_sphere_t)
            h1t_d = treams.special.spherical_hankel1_d(l, x_sphere_t)
            f23 = psi**2 * h1t
            d23 = psi**2 * j1t

    if np.abs(mat_sphere[2]) == 0 and np.abs(mat_env[2]) == 0:
        delta = x_sphere * mat_env[0] / (x_env * mat_sphere[0]) 
        matrix = np.zeros((2, 2), complex)
        rhs = np.zeros(2, complex)
        matrix[0][0] = -d12
        matrix[0][1] = (d14 + tm[0] * f14) / delta
        rhs[0] = d1L
        matrix[1][0] = -d22 / x_env
        matrix[1][1] = (d24 + tm[0] * f24) / x_sphere
        rhs[1] = d2L / x_env
        sol = np.linalg.solve(matrix, rhs)
        res[0] = sol[0]
        return res
    elif np.abs(mat_sphere[2]) == 0 and np.abs(mat_env[2]) > 0:
        d44 = -mat_sphere[0] / mat_env[0] * 0.5 * x_env_t**2 * j1 
        f44 = -mat_sphere[0] / mat_env[0] * 0.5 * x_env_t**2 * h1 
        if l == 0:
            matrix = np.zeros((2, 2), complex)
            rhs = np.zeros(2, complex)
            matrix[0][0] = -d22 / x_env
            matrix[0][1] = (d24 + tm[0] * f24) / x_sphere
            matrix[1][0] = -d42 / x_env
            matrix[1][1] = (d44 + tm[0] * f44) / x_sphere
            rhs[0] = d2L / x_env
            rhs[1] = d4L / x_env
            sol = np.linalg.solve(matrix, rhs)
            res[0] = sol[0]
            return res
        elif l > 0:
            matrix = np.zeros((3, 3), complex)
            rhs_L = np.zeros(3, complex)
            rhs_N = np.zeros(3, complex)
            matrix[0][0] = d21 / x_env_t
            matrix[0][1] = -psi * d22 / x_env
            matrix[0][2] = psi * (d24 + tm[0] * f24) / x_sphere
            rhs_L[0] = psi * d2L / x_env
            rhs_L[1] = psi * d3L / x_env
            rhs_L[2] = psi * d4L / x_env
            matrix[1][0] = d31 / x_env_t
            matrix[1][1] = -psi * d32 / x_env
            matrix[2][0] = d41 / x_env_t
            matrix[2][1] = -psi * d42 / x_env
            matrix[2][2] = psi * (d44 + tm[0] * f44) / x_sphere
            sol_L = np.linalg.solve(matrix, rhs_L)
            res[0] = sol_L[1]
            res[1] = sol_L[0]
            rhs_N[0] = -d2N / x_env_t
            rhs_N[1] = -d3N / x_env_t
            rhs_N[2] = -d4N / x_env_t
            sol_N = np.linalg.solve(matrix, rhs_N)
            res[2] = sol_N[1]
            res[3] = sol_N[0]
            return res
    elif np.abs(mat_sphere[2]) > 0 and np.abs(mat_env[2]) == 0:
        d42 = -mat_env[0]/mat_sphere[0] * x_sphere_t**2 * 0.5 * h
        d44 = (l * (l + 1) - 0.5 * x_sphere_t**2) * j1 - 2 * x_sphere * j1_d
        f44 = (l * (l + 1) - 0.5 * x_sphere_t**2) * h1 - 2 * x_sphere * h1_d
        d4L = -mat_env[0]/mat_sphere[0] * 0.5 * x_sphere_t**2 * j
        if l == 0:
            matrix = np.zeros((2, 2), complex)
            rhs = np.zeros(2, complex)
            matrix[0][0] = -d22 / x_env
            matrix[0][1] = (d24 + tm[0] * f24) / x_sphere
            matrix[1][0] = -d42 / x_env
            matrix[1][1] = (d44 + tm[0] * f44) / x_sphere
            rhs[0] = d2L / x_env
            rhs[1] = d4L / x_env
            sol = np.linalg.solve(matrix, rhs)
            res[0] = sol[0]
            return res
        elif l > 0:
            d33 = (l * (l + 1) - 0.5 * x_sphere_t**2 - 1) * j1t - x_sphere_t * j1t_d
            d43 = l * (l + 1) * (x_sphere_t * j1t_d - j1t)
            f33 = (psi**2 - 0.5 * x_sphere_t**2 - 1) * h1t - x_sphere_t * h1t_d
            f43 = psi**2 * (x_sphere_t * h1t_d - h1t)  
            d34 = x_sphere * j1_d - j1    
            f34 = x_sphere * h1_d - h1 
            matrix = np.zeros((3, 3), complex)
            rhs = np.zeros(3, complex)
            matrix[0][0] = -psi * d22 / x_env
            matrix[2][0] = -psi * d42 / x_env
            matrix[0][1] = -d23 / x_sphere_t - f23 * tm[3] / x_sphere_t + psi * tm[2] * f24 / x_sphere
            matrix[1][1] = -d33 / x_sphere_t - f33 * tm[3] / x_sphere_t + psi * tm[2] * f34 / x_sphere
            matrix[2][1] = -d43 / x_sphere_t - f43 * tm[3] / x_sphere_t + psi * tm[2] * f44 / x_sphere
            matrix[0][2] = psi * (d24 + tm[0] * f24) / x_sphere - tm[1] * f23 / x_sphere_t
            matrix[1][2] = psi * (d34 + tm[0] * f34) / x_sphere - tm[1] * f33 / x_sphere_t
            matrix[2][2] = psi * (d44 + tm[0] * f44) / x_sphere - tm[1] * f43 / x_sphere_t 
            rhs[0] = psi * d2L / x_env
            rhs[2] = psi * d4L / x_env
            sol = np.linalg.solve(matrix, rhs)
            res[0] = sol[0]
            return res 
    elif np.abs(mat_sphere[2]) > 0 and np.abs(mat_env[2]) > 0:
        d44 = (mat_sphere[0]/mat_env[0] * (x_env_t/x_sphere_t)**2 
               * ((psi**2 - 0.5 * x_sphere_t**2) * j1 - 2 * x_sphere * j1_d))
        f44 = (mat_sphere[0]/mat_env[0] * (x_env_t/x_sphere_t)**2 
               * ((psi**2 - 0.5 * x_sphere_t**2) * h1 - 2 * x_sphere * h1_d))
        if l == 0:
            matrix = np.zeros((2, 2), complex)
            rhs_L = np.zeros(2, complex)
            matrix[0][0] = -d22 / x_env
            matrix[1][0] = -d42 / x_env
            matrix[0][1] = (d24 + tm[0] * f24) / x_sphere
            matrix[1][1] = (d44 + tm[0] * f44) / x_sphere
            rhs_L[0] = d2L / x_env
            rhs_L[1] = d4L / x_env
            sol_L = np.linalg.solve(matrix, rhs_L)
            res[0] = sol_L[0]
            return res
        elif l > 0:
            d11 = x_env_t * ht_d + ht
            d13 = x_sphere_t * j1t_d + j1t
            d33 = (mat_sphere[0]/mat_env[0] * (x_env_t/x_sphere_t)**2 
                * ((psi**2 - 0.5 * x_sphere_t**2 - 1) * j1t - x_sphere_t * j1t_d))
            d43 = (mat_sphere[0]/mat_env[0] * (x_env_t/x_sphere_t)**2  * psi**2 
                * (x_sphere_t * j1t_d - j1t))
            d34 = (mat_sphere[0]/mat_env[0] * (x_env_t/x_sphere_t)**2 
                * (x_sphere * j1_d - j1))
            f34 = (mat_sphere[0]/mat_env[0] * (x_env_t/x_sphere_t)**2 
               * (x_sphere * h1_d - h1))
            f13 = x_sphere_t * h1t_d + h1t
            f33 = (mat_sphere[0]/mat_env[0] * (x_env_t/x_sphere_t)**2 
               * ((psi**2 - 0.5 * x_sphere_t**2 - 1) * h1t - x_sphere_t * h1t_d))
            f43 = (mat_sphere[0]/mat_env[0] * (x_env_t/x_sphere_t)**2  * psi**2 
               * (x_sphere_t * h1t_d - h1t))
            d1N = x_env_t * jt_d + jt
            matrix = np.zeros((4, 4), complex)
            rhs_L = np.zeros(4, complex)
            rhs_N = np.zeros(4, complex)
            matrix[0][0] = d11 / x_env_t
            matrix[1][0] = d21 / x_env_t
            matrix[2][0] = d31 / x_env_t
            matrix[3][0] = d41 / x_env_t
            matrix[0][1] = -psi * d12 / x_env
            matrix[1][1] = -psi * d22 / x_env
            matrix[2][1] = -psi * d32 / x_env
            matrix[3][1] = -psi * d42 / x_env
            matrix[0][2] = -d13 / x_sphere_t - tm[3] * f13 / x_sphere_t + psi * tm[2] * f14 / x_sphere
            matrix[1][2] = -d23 / x_sphere_t - tm[3] * f23 / x_sphere_t + psi * tm[2] * f24 / x_sphere
            matrix[2][2] = -d33 / x_sphere_t - tm[3] * f33 / x_sphere_t + psi * tm[2] * f34 / x_sphere
            matrix[3][2] = -d43 / x_sphere_t - tm[3] * f43 / x_sphere_t + psi * tm[2] * f44 / x_sphere
            matrix[0][3] = psi * (d14 + tm[0] * f14) / x_sphere - tm[1] * f13 / x_sphere_t 
            matrix[1][3] = psi * (d24 + tm[0] * f24) / x_sphere - tm[1] * f23 / x_sphere_t 
            matrix[2][3] = psi * (d34 + tm[0] * f34) / x_sphere - tm[1] * f33 / x_sphere_t 
            matrix[3][3] = psi * (d44 + tm[0] * f44) / x_sphere - tm[1] * f43 / x_sphere_t 
            rhs_L[0] = psi * d1L / x_env
            rhs_L[1] = psi * d2L / x_env
            rhs_L[2] = psi * d3L / x_env
            rhs_L[3] = psi * d4L / x_env
            rhs_N[0] = -d1N / x_env_t
            rhs_N[1] = -d2N / x_env_t
            rhs_N[2] = -d3N / x_env_t
            rhs_N[3] = -d4N / x_env_t
            sol_N = np.linalg.solve(matrix, rhs_N)
            res[2] = sol_N[1]
            res[3] = sol_N[0] 
            sol_L = np.linalg.solve(matrix, rhs_L)
            res[0] = sol_L[1]
            res[1] = sol_L[0] 
            return res
    return res 

def mie_acoustics(l, x, *materials):
    r"""Mie-scattering coefficient of degree :math:`l` for a sphere.

    The sphere is defined by its size parameter :math:`k_0 r`, where :math:`r` is the
    radius and :math:`k_0` the wavenumber in the air. A multilayered sphere is defined
    by an array of ascending numbers that define the size parameters of the layers
    starting from the center.

    Likewise, the material parameters are given from inside to outside. These arrays
    are expected to contain exactly one element more than the array `x`.

    The result is a complex number relating the incident and scattered field coefficients,
    both indexed identically.

    Note:
        1. :math:`c_t` of the **last** material must be zero. 
        2. For the soft and hard spheres, only one radius must be given. 

    Args:
        l (int or array_like): Degree :math:`l \geq 0`.
        x (float or array_like): Size parameters in the array.
        rho (array_like of float or complex): Mass density (kg/m^3). The length must be at least two.
        c (array_like of float or complex): Longitudinal speed of sound (m/s). The length must be at least two.
        c_t (array_like of float or complex): Transverse speed of sound (m/s). The length must be at least two.

    Returns:
        complex array
    """
    
    mat = list(zip(*materials))
    l = np.atleast_1d(l)
    x = np.atleast_1d(x)
    res = []
    for j in l:
        mie = np.zeros(4, complex)
        for i in range(len(list(mat)) - 1):
            mat_sphere, mat_env = mat[i], mat[i + 1]
            mie = _mie_acoustics_iter(mie, j, x[i], mat_sphere, mat_env)
        res.append(mie[0])
    return res

def _mie_acoustics_iter_cyl(tm, kz, m, k0, radius, mat_cyl, mat_env):
    """Solve a matrix for the interface conditions at a cylindrical interface.

    Three types of interfaces are supported: fluid-fluid, soft-fluid, and hard-fluid.

    The fluid medium has :math:`(\rho, c, c_t = 0)`,
    the soft medium has :math:`(\rho = 0, c = 0, c_t = 0)`,
    and the hard medium has :math:`(\rho = \infty, c = 0, c_t = 0)`.

    Args:
        tm (complex): T-matrix of the previous layer.
        kz (float): Z-component of the wave vector.
        m (int): Order.
        k0 (float): Angular wavenumber in the air.
        radius (float): Radius of the cylinder.
        mat_cyl (array_like of float or complex): Material of the cylinder :math:`(\rho, c, c_t)`.
        mat_env (array_like of float or complex): Material of the medium :math:`(\rho, c, c_t)`.

    Returns:
        complex
    """ 

    k_env = k0 * AcousticMaterial().c / mat_env[1] + 0j
    krho_env = np.sqrt(k_env * k_env - kz * kz)
    x_env = krho_env * radius
    j = treams.special.jv(m, x_env)
    j_d = treams.special.jv_d(m, x_env)
    h = treams.special.hankel1(m, x_env)
    h_d = treams.special.hankel1_d(m, x_env)
    res = 0
    if np.abs(mat_cyl[2]) == 0 and np.abs(mat_cyl[1]) == 0 and np.abs(mat_cyl[0]) == np.inf:
        return -j_d / h_d
    elif np.abs(mat_cyl[2]) == 0 and np.abs(mat_cyl[1]) == 0 and np.abs(mat_cyl[0]) == 0:
        return -j / h
    elif np.abs(mat_cyl[1]) > 0 and np.abs(mat_cyl[2]) == 0:
        k_cyl = k0 * AcousticMaterial().c / mat_cyl[1] + 0j
        krho_cyl = np.sqrt(k_cyl * k_cyl - kz * kz)
        x_cyl = krho_cyl * radius
        j1 = treams.special.jv(m, x_cyl)
        j1_d = treams.special.jv_d(m, x_cyl)
        h1 = treams.special.hankel1(m, x_cyl)
        h1_d = treams.special.hankel1_d(m, x_cyl)
        matrix = np.zeros((2, 2), complex)
        rhs = np.zeros(2, complex)
        rhs[0] = x_env * j_d
        rhs[1] = j
        aL = x_cyl * (j1_d + tm * h1_d) * mat_env[0]/mat_cyl[0]
        cL = j1 + tm * h1
        al = -x_env * h_d
        cl = -h
        matrix[0][0] = aL
        matrix[1][0] = cL
        matrix[0][1] = al
        matrix[1][1] = cl
        res = np.linalg.solve(matrix, rhs)
        return res[1]
    return res


def mie_acoustics_cyl(kz, m, k0, radii, *materials):
    r"""Scattering coefficient at a cylindrical interface

    The cylinder is defined by its radii. A multilayered cylinder is defined
    by giving an array of ascending numbers that define the size parameters of the
    layers starting from the center.

    Likewise, the material parameters are given from inside to outside. These arrays
    are expected to contain exactly one element more than the array `radii`.

    The result is a complex number relating the incident and scattered field 
    coefficients, both indexed identically.

    Note:
        1. :math:`c_t` of **all** the materials must be zero. 
        2. For the soft and hard infinite cylinders, only one radius must be given. 

    Args:
        kz (float): Z-component of the wave vector in the medium.
        m (int or array_like): Order.
        k0 (float or complex): Angular wavenumber in the air. Has units of :math:`kz`.
        radii (float or array_like): Radii of the layers. Has units of :math:`1/kz`.
        rho (float or complex or array_like): Mass density (kg/m^3).
        c (float or complex or array_like): Longitudinal speed of sound (m/s).
        c_t (float or complex or array_like): Transverse speed of sound (m/s).

    Returns:
        complex array
    """
    
    mat = list(zip(*materials))
    radii = np.atleast_1d(radii)
    m = np.atleast_1d(m)
    res = []
    for j in m:
        mie = 0
        for i in range(len(list(mat)) - 1):
            mat_cyl, mat_env = mat[i], mat[i + 1]
            mie = _mie_acoustics_iter_cyl(mie, kz, j, k0, radii[i], mat_cyl, mat_env)
        res.append(mie)
    return res

def fresnel_acoustics(kzs, rhos):
    r"""Fresnel coefficients for a planar interface.

    The first two dimensions index the media above and below the S-matrix, 
    the second two dimensions were added to meet the acoustotreams convention.

    The result is an array relating the incoming and the outgoing mode coefficients, 
    both indexed identically. The first dimension of the array corresponds to 
    the outgoing modes and the second one to the incoming modes.

    Args:
        kzs (float): Z-components of the wave vector (rad/m).
        rhos (float or complex): Mass densities (kg/m^3).

    Returns:
        complex (2, 2, 1, 1)-array
    """
    res = np.zeros((2, 2, 1, 1), complex)
    res[1][1][0][0] = 2 * rhos[0] * kzs[1] / (rhos[0] * kzs[1] + rhos[1] * kzs[0])
    res[0][0][0][0] = 2 * rhos[1] * kzs[0] / (rhos[0] * kzs[1] + rhos[1] * kzs[0])
    res[1][0][0][0] = (-rhos[0] * kzs[1] + rhos[1] * kzs[0]) / (rhos[0] * kzs[1] + rhos[1] * kzs[0])
    res[0][1][0][0] = -res[1][0][0][0]
    return res


import numpy as np
import cmath

from treams import misc

class AcousticMaterial:
    r"""Material definition.

    The material properties are defined in the frequency domain by scalar values
    for density :math:`\rho`, the speed of longitudinal elastic waves :math:`c`, and
    the speed of transverse elastic waves :math:`c_t`. Materials are, thus, assumed to be linear, 
    time-invariant, homogeneous, isotropic, local and to have no gain.

    Args:
        rho (complex, optional): Mass density (kg/m^3). Defaults to 1.3 kg/m^3.
        c (complex, optional): Longitudinal speed of sound (m/s). Defaults to 343 m/s.
        ct (complex, optional): Transverse speed of sound (m/s). Defaults to 0 m/s.
    """

    def __init__(self, rho=1.3, c=343., ct=0.):
        """Initialization."""
        if isinstance(rho, AcousticMaterial):
            rho, c, ct = rho()
        elif isinstance(rho, (tuple, list, np.ndarray)):
            if len(rho) == 0:
               rho = 1.3
            elif len(rho) == 1:
                rho = rho[0]
            elif len(rho) == 2:
                rho, c = rho
            elif len(rho) == 3:
                rho, c, ct = rho
            else:
                raise ValueError("invalid material definition")
        self._rho = rho
        self._c = c
        self._ct = ct

    @property
    def rho(self):
        """Mass density.

        Returns:
            float or complex
        """
        return self._rho

    @property
    def c(self):
        """Speed of longitudinal elastic waves.

        Returns:
            float or complex
        """
        return self._c
    
    @property
    def ct(self):
        """Speed of transverse elastic waves.

        Returns:
            float or complex
        """
        return self._ct
    
    def __iter__(self):
        """Iterator for a tuple containing the material parameters.

        Useful for unpacking the material parameters into a function that takes these
        parameters separately, e.g. ``foo(*material)``.
        """
        return iter((self.rho, self.c, self.ct))
    
    @classmethod
    def from_params(cls, rho=1.3, params=(151767.21, 0.)):
        r"""Create acoustic material from Lamé parameters.

        This function calculates the longitudinal and transverse speeds of sound with
        :math:`c = \sqrt{\frac{\lambda+2\mu}{\rho}}` and :math:`c_t = \sqrt{\frac{\mu}{\rho}}`. 

        Args:
            rho (complex, optional): Mass density (kg/m^3). Defaults to 1.3 kg/m^3.
            params (complex, optional): Lamé parameters (Pa). Defaults to (151767.21, 0.) Pa.

        Returns:
            AcousticMaterial
        """
        c = cmath.sqrt((params[0] + 2 * params[1]) / rho)
        ct = cmath.sqrt(params[1] / rho)
        return cls(rho, c, ct)
    
    @classmethod
    def from_pratio(cls, rho=1.3, c=343, pratio=0.5):
        r"""Create acoustic material from Poisson's ratio.

        This function calculates the transverse speed of sound with
        :math:`c_t = c \sqrt{\frac{1 - 2\sigma}{2 - 2\sigma}}`. 

        Args:
            rho (complex, optional): Mass density (kg/m^3). Defaults to 1.3 kg/m^3.
            c (complex, optional): Longitudinal speed of sound (m/s). Defaults to 343 m/s.
            pratio (complex, optional): Poisson's ratio. Defaults to 0.5.

        Returns:
            AcousticMaterial
        """
        ct = c * cmath.sqrt((1 - 2 * pratio) / (2 - 2 * pratio))
        return cls(rho, c, ct)
    
    @property
    def impedance(self):
        r"""Relative impedance for longitudinal waves (Pa*s/m^3).

        The relative impedance is defined by :math:`Z = \rho c`.

        Returns:
            complex
        """
        return self.rho * self.c
    
    @property
    def impedancet(self):
        r"""Relative impedance for transverse waves (Pa*s/m^3).

        The relative impedance is defined by :math:`Z = \rho c_t`.

        Returns:
            complex
        """
        return self.rho * self.ct
    
    def __call__(self):
        """Return a tuple containing all material parameters.

        Returns:
            tuple
        """
        return self.rho, self.c, self.ct

    def __eq__(self, other):
        """Compare material parameters.

        Materials are considered equal if all material parameters are equal.
        Moreover, compares objects that contain at most three values.

        Returns:
            bool
        """
        if other is None:
            return False
        if not isinstance(other, AcousticMaterial):
            other = AcousticMaterial(*other)
        return (
            self.rho == other.rho
            and self.c == other.c
            and self.ct == other.ct
        )
    
    @property
    def isreal(self):
        """Test if the material has purely real-valued parameters.

        Returns:
            bool
        """
        return all(i.imag == 0 for i in self)
    
    @property
    def isshear(self):
        """Test if the material sustains shear waves.

        Returns:
            bool
        """
        return self.ct != 0

    def __str__(self):
        """All three material parameters.

        Returns:
            str
        """
        return "(" + ", ".join([str(i) for i in self()]) + ")"

    def __repr__(self):
        """Representation that allows recreating the object.

        Returns:
            str
        """
        return self.__class__.__name__ + str(self)

    def ks(self, k0):
        """Return the wavenumber of longitudinal waves in the medium.

        Args:
            k0 (float): Angular wavenumber in the air.

        Returns:
            tuple
        """
        if isinstance(self.c, (tuple, list, np.ndarray)):
            return tuple(k0 * AcousticMaterial().c / c for c in self.c)
        else:
            return k0 * AcousticMaterial().c / self.c

        
    def kst(self, k0):
        """Return the wavenumber of transverse waves in the medium.

        Args:
            k0 (float): Angular wavenumber in the air.

        Returns:
            tuple
        """
        if isinstance(self.ct, (tuple, list, np.ndarray)):
            return tuple(k0 * AcousticMaterial().c / ct for ct in self.ct)
        else:
            return k0 * AcousticMaterial().c / self.ct
    
    def krhos(self, k0, kz):
        r"""The (cylindrically) radial part of the wave vector for longitudinal waves .

        The cylindrically radial part is defined by :math:`k_\rho = \sqrt(k^2 - k_z^2)`.
        The returned values have non-negative imaginary parts.

        Args:
            k0 (float): Angular wavenumber in the air.
            kz (float or array_like): Z-component of the wave vector. Has units of k0.

        Returns:
            complex or array_like
        """
        ks = self.ks(k0)
        return misc.wave_vec_z(kz, 0, ks)

    def kzs(self, k0, kx, ky):
        r"""The z-component of the wave vector for longitudinal waves.

        The z-component of the wave vector is defined by
        :math:`k_z = \sqrt(k^2 - k_x^2 - k_y^2)`. The returned values have
        non-negative imaginary parts.

        Args:
            k0 (float): Angular wavenumber in the air.
            kx (float or array_like): X-component of the wave vector. Has units of k0.
            ky (float or array_like): Y-component of the wave vector. Has units of k0.

        Returns:
            complex or array_like
        """
        ks = self.ks(k0)
        return misc.wave_vec_z(kx, ky, ks)   
    
    def krhost(self, k0, kz):
        r"""The (cylindrically) radial part of the wave vector for transverse waves.

        The cylindrically radial part is defined by :math:`k_\rho = \sqrt(k^2 - k_z^2)`.
        The returned values have non-negative imaginary parts.

        Args:
            k0 (float): Angular wavenumber in the air.
            kz (float or array_like): Z-component of the wave vector. Has units of k0.

        Returns:
            complex or array_like
        """
        ks = self.kst(k0)
        return misc.wave_vec_z(kz, 0, ks)

    def kzst(self, k0, kx, ky):
        r"""The z-component of the wave vector for transverse waves.

        The z-component of the wave vector is defined by
        :math:`k_z = \sqrt(k^2 - k_x^2 - k_y^2)`. The returned values have
        non-negative imaginary parts.

        Args:
            k0 (float): Angular wavenumber in the air.
            kx (float or array_like): X-component of the wave vector. Has units of k0.
            ky (float or array_like): Y-component of the wave vector. Has units of k0.

        Returns:
            complex or array_like
        """
        ks = self.kst(k0)
        return misc.wave_vec_z(kx, ky, ks) 


r"""Special (mathematical) functions.

.. currentmodule:: acoustotreams.special

Special mathematical functions used in :mod:`acoustotreams`. Some functions are imported from
:py:mod:`scipy.special` and :py:mod:`treams.special`.

Available functions
===================

Spherical waves
---------------

.. autosummary::
   :toctree: generated/

   ssw_Psi
   ssw_psi
   vsw_L
   ssw_rPsi
   vsw_rL
   tl_ssw
   tl_ssw_r

Cylindrical waves
-----------------

.. autosummary::
   :toctree: generated/

   scw_Psi
   scw_psi
   vcw_L
   scw_rPsi
   vcw_rL
   tl_scw
   tl_scw_r


Plane waves
-----------

.. autosummary::
   :toctree: generated/

   spw_Psi
   vpw_L


Those functions are just imported from Scipy. So, one only needs to import this
package within acoustotreams.

+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.hankel1`\(v, z[, out])            | Hankel function of the  |
|                                                            | first kind.             |
+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.hankel2`\(v, z[, out])            | Hankel function of the  |
|                                                            | second kind.            |
+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.jv`\(v, z[, out])                 | Bessel function of the  |
|                                                            | first kind of real      |
|                                                            | order and complex       |
|                                                            | argument.               |
+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.yv`\(v, z[, out])                 | Bessel function of the  |
|                                                            | second kind of real     |
|                                                            | order and complex       |
|                                                            | argument.               |
+------------------------------------------------------------+-------------------------+
| | :py:func:`spherical_jn <scipy.special.spherical_jn>`\(n, | Spherical Bessel        |
|   z[, derivative])                                         | function of the first   |
|                                                            | kind or its derivative. |
+------------------------------------------------------------+-------------------------+
| | :py:func:`spherical_yn <scipy.special.spherical_yn>`\(n, | Spherical Bessel        |
|   z[, derivative])                                         | function of the second  |
|                                                            | kind or its derivative. |
+------------------------------------------------------------+-------------------------+

The following functions are just imported from treams. So, one only needs to import this
package within treams.

Bessel and Hankel functions, with their spherical counterparts, derivatives
---------------------------------------------------------------------------

:py:data:`~treams.special.hankel1_d`
:py:data:`~treams.special.hankel2_d`
:py:data:`~treams.special.jv_d`
:py:data:`~treams.special.yv_d`
:py:data:`~treams.special.spherical_hankel1`
:py:data:`~treams.special.spherical_hankel2`
:py:data:`~treams.special.spherical_hankel1_d`
:py:data:`~treams.special.spherical_hankel2_d`

Those functions just wrap Scipy functions with special optional arguments to be able to
analogously access them like their non-spherical counterparts:

:py:data:`~treams.special.spherical_jn_d`
:py:data:`~treams.special.spherical_yn_d`


Scipy functions with enhanced domain
------------------------------------

:py:data:`~treams.special.sph_harm`
:py:data:`~treams.special.lpmv`


Integrals for the Ewald summation
---------------------------------

:py:data:`~treams.special.incgamma`
:py:data:`~treams.special.intkambe`


Wigner d- and Wigner D-matrix elements
--------------------------------------

:py:data:`~treams.special.wignersmalld`
:py:data:`~treams.special.wignerd`


Wigner 3j-symbols
-----------------

:py:data:`~treams.special.wigner3j`

Vector spherical harmonics
--------------------------
   
:py:data:`~treams.special.pi_fun`
:py:data:`~treams.special.tau_fun`
:py:data:`~treams.special.vsh_X`
:py:data:`~treams.special.vsh_Y`
:py:data:`~treams.special.vsh_Z`


Coordinate system transformations
---------------------------------

:py:data:`~treams.special.car2cyl`
:py:data:`~treams.special.car2sph`
:py:data:`~treams.special.cyl2car`
:py:data:`~treams.special.cyl2sph`
:py:data:`~treams.special.sph2car`
:py:data:`~treams.special.sph2cyl`
:py:data:`~treams.special.vcar2cyl`
:py:data:`~treams.special.vcar2sph`
:py:data:`~treams.special.vcyl2car`
:py:data:`~treams.special.vcyl2sph`
:py:data:`~treams.special.vsph2car`
:py:data:`~treams.special.vsph2cyl`
:py:data:`~treams.special.car2pol`
:py:data:`~treams.special.pol2car`
:py:data:`~treams.special.vcar2pol`
:py:data:`~treams.special.vpol2car`

"""

from scipy.special import (  # noqa: F401
    hankel1,
    hankel2,
    jv,
    yv,
    spherical_jn,
    spherical_yn,
)

from treams.special import(   # noqa: F401
    spherical_jn_d,
    spherical_yn_d,
    sph_harm,
    lpmv,
    incgamma,
    intkambe,
    wignersmalld,
    wignerd,
    wigner3j,
    pi_fun,
    tau_fun,
    vsh_X,
    vsh_Y,
    vsh_Z,
    car2cyl,
    car2sph,
    cyl2car,
    cyl2sph,
    sph2car,
    sph2cyl,
    vcar2cyl,
    vcar2sph,
    vcyl2car,
    vcyl2sph,
    vsph2car,
    vsph2cyl,
    car2pol,
    pol2car,
    vcar2pol,
    vpol2car,
    )

from acoustotreams.special._wavesacoustics import *  # noqa: F401


"""Scalar spherical wave module.

.. autosummary::
   :toctree:

   periodic_to_scw
   periodic_to_spw
   rotate
   translate
   translate_periodic

"""

import acoustotreams.special as ats
import treams.special as sc
import numpy as np
from treams import lattice
import scipy.special as ss

def _translate_s(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs):
    """Regular translation coefficient for spherical modes"""
    if abs(kr) < 1e-16:
        return 0.0j
    return ats.tl_ssw(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs)

def _translate_r(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs):
    """Singular translation coefficient for spherical modes"""
    return ats.tl_ssw_r(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs)


_translate_s = np.vectorize(_translate_s)
_translate_r = np.vectorize(_translate_r)


def translate(lambda_, mu, l, m, kr, theta, phi, singular=True, *args, **kwargs):
    """translate(lambda_, mu, l, m, kr, theta, phi, singular=True)

    Translation coefficients for spherical waves.

    Returns the correct translation coefficients from :func:`acoustotreams.ssw.tl_ssw`
    and :func:`acoustotreams.ssw.tl_ssw_r` or combinations thereof for the specified modes and
    basis.

    Args:
        lambda_ (int or array_like): Degree of output modes.
        mu (int or array_like): Order of output modes.
        l (int or array_like): Degree of input modes.
        m (int or array_like): Order of input modes.
        kr (float or complex or array_like): Translation distance in units of the wavenumber.
        theta (float or array_like): Polar angle (rad).
        phi (float or array_like): Azimuthal angle (rad).
        singular (bool, optional): If True, singular translation coefficients are used,
            else regular coefficients. Defaults to ``True``.

    Returns:
        complex or array_like
    """
    if singular:
        return _translate_s(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs)
    return _translate_r(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs)


def _rotate(lambda_, mu, l, m, phi, theta, psi, *args, **kwargs):
    """
    Rotation coefficients for the rotation by the Euler angles phi, theta, and psi.
    It is intended to be used for constructing the rotation matrix manually.
    """
    if lambda_ == l:
        return sc.wignerd(l, mu, m, phi, theta, psi, *args, **kwargs)
    else:
        return 0.0j

_rotate = np.vectorize(_rotate)


def rotate(lambda_, mu, l, m, phi, theta=0, psi=0, *args, **kwargs):
    """rotate(lambda_, mu, l, m, phi, theta=0, psi=0)
    
    Rotation coefficients for scalar spherical modes.

    Return the correct rotation coefficients from :func:`treams.special.wignerd`. The
    angles are given as the Euler angles in `z-y-z`-convention. In the intrinsic (object
    fixed coordinate system) convention, the rotations are applied in the order phi
    first, theta second, and psi third. In the extrinsic (global or reference frame fixed
    coordinate system), the rotations are applied psi first, theta second, phi third.

    Args:
        lambda_ (int or array_like): Degree of output modes.
        mu (int or array_like): Order of output modes.
        l (int or array_like): Degree of input modes.
        m (int or array_like): Order of input modes.
        phi (float or complex or array_like): First Euler angle.
        theta (float or array_like): Second Euler angle.
        psi (float or array_like): Third Euler angle.

    Returns:
        complex or array_like
    """
    return _rotate(lambda_, mu, l, m, phi, theta, psi, *args, **kwargs)


def _transl_a_lattice(lambda_, mu, l, m, dlms):
    """Singular translation coefficient for a scalar spherical wave on a lattice"""
    pref = np.power(-1, np.abs(m)) * np.sqrt(4 * np.pi * (2 * l + 1) * (2 * lambda_ + 1)) * np.power(1.0j, lambda_ - l)
    dlm = 0
    res = 0
    for p in range(l + lambda_, max(abs(lambda_ - l), abs(mu - m)) - 1, -2):
        dlm = dlms[p * (p + 1) + m - mu]
        res += dlm * np.power(1.0j, p) * np.sqrt(2 * p + 1) * sc.wigner3j(l, lambda_, p, m, -mu, -m + mu) * sc.wigner3j(l, lambda_, p, 0, 0, 0)
    return res * pref


def translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, eta=0, func=lattice.lsumsw):
    """Translation coefficients for scalar spherical waves on a lattice.

    Return the translation coefficents for the given modes on a lattice. The computations
    employ the fast converging Ewald summation from :mod:`treams.lattice`.


    Args:
        ks (float or complex): Wavenumber of longitudinal waves in the medium.
        kpar ((D,)-array of float): Parallel component of the wave vector. 
            Defines the dimension with `1 <= D <= 3`. Has units of ks.
        a ((D,D)-array of float): Lattice vectors in each row of the array.
        rs ((M, 3)-array of float): Shift vectors with respect to one lattice point. 
            Have units of 1/ks.
        out (2- or 3-tuple of integer arrays): Output modes.
        in_ (2- or 3-tuple of integer arrays): Input modes. If None is given, they are equal to
            the output modes.
        rsin (float): Shift vectors of the input modes. If None is given, they are equal to `rs`.
            Have units of 1/ks.
        eta (float or complex, optional): Splitting parameter for the Ewald summation of
            lattice sums. If zero, an estimation for the optimal value is done. Defaults
            to zero.

    Returns:
        array_like of complex
    """
    if in_ is None:
        in_ = out
    out = (*(np.array(o) for o in out),)
    in_ = (*(np.array(i) for i in in_),)
    if len(out) < 2 or len(out) > 3:
        raise ValueError(f"invalid length of output modes {len(out)}, must be 2 or 3")
    elif len(out) == 2:
        out = (np.zeros_like(out[0]),) + out
    if len(in_) < 2 or len(in_) > 3:
        raise ValueError(f"invalid length of input modes {len(in_)}, must be 2 or 3")
    elif len(in_) == 2:
        in_ = (np.zeros_like(in_[0]),) + in_
    if rsin is None:
        rsin = rs
    modes = np.array([
        [l, m]
        for l in range(np.max(out[1]) + np.max(in_[1]) + 1)
        for m in range(-l, l + 1)
    ])
    kpar = np.array(kpar)
    rs = np.array(rs)
    if rs.ndim == 1:
        rs = rs.reshape((1, -1))
    rsin = np.array(rsin)
    if rsin.ndim == 1:
        rsin = rsin.reshape((1, -1))
    rsdiff = -rs[:, None, None, :] + rsin[:, None, :]

    dim = 1 if kpar.ndim == 0 else kpar.shape[-1]
    # The result has the shape (n_rs, n_rs, n_modes)
    dlms = func(dim, modes[:, 0], modes[:, 1], ks, kpar, a, rsdiff, eta)
    res = np.zeros((out[1:][0].shape[-1], in_[1:][0].shape[-1]), complex)
    for i in range(out[1:][0].shape[-1]):
        for j in range(in_[1:][0].shape[-1]):
                res[i][j] = _transl_a_lattice(out[1:][0][i], out[1:][1][i], in_[1:][0][j], in_[1:][1][j], dlms[out[0][i]][in_[0][j]])
    return res
         

def _periodic_to_spw(kx, ky, kz, l, m, area, *args, **kwargs):
    """Convert a periodic spherical wave into plane waves"""
    k = np.sqrt(kx * kx + ky * ky + kz * kz)
    phi = np.arctan2(ky, kx)
    kz_s = kz
    if kz == 0:
        kz_s = 1e-20 + 1e-20j
    elif (np.imag(kz_s) < 0) or (np.imag(kz_s) == 0 and np.real(kz_s) < 0):
        kz_s = -kz_s
    prefactor = (
        np.sqrt(np.pi * (2 * l + 1))
        * np.sqrt(ss.gamma(l - m + 1) / ss.gamma(l + m + 1))
        * np.exp(1j * m * phi)
        * np.power(-1j, l)
        / (area * k * kz_s)
    )
    return prefactor * sc.lpmv(m, l, kz / k + 0j, *args, **kwargs)

_periodic_to_spw = np.vectorize(_periodic_to_spw)


def periodic_to_spw(kx, ky, kz, l, m, area, *args, **kwargs):
    """periodic_to_spw(kx, ky, kz, l, m, area)
    
    Convert periodic scalar spherical waves into plane waves.

    Return the coefficient for the basis change from spherical waves on a lattice
    to plane waves. For multiple positions, only the diagonal values (corresponding to
    identical positions) are returned. A correct phase factor is still necessary for the correct
    result.

    Args:
        kx (float or array_like): X-component of the wave vector of plane waves.
        ky (float or array_like): Y-component of the wave vector of plane waves.
        kz (float or complex or array_like): Z-component of the wave vector of plane waves.
        l (int or array_like): Degree of spherical waves.
        m (int or array_like): Order of spherical waves.
        area (float or array_like): Unit cell area.

    Returns:
        complex
    """
    return _periodic_to_spw(kx, ky, kz, l, m, area, *args, **kwargs)

def _periodic_to_scw(kz, m, l, mu, k, a, *args, **kwargs):
    """Convert periodic spherical wave to cylindrical wave"""
    if mu != m:
        return 0.0j + sc.lpmv(0, 1, 0, *args, **kwargs)
    prefactor = (
        np.sqrt(np.pi * (2 * l + 1), *args, **kwargs) *
        np.sqrt(ss.gamma(l - m + 1) / ss.gamma(l + m + 1), *args, **kwargs) * 
        0.5 * 
        np.power(-1j, l - m, *args, **kwargs)
        / (a * k)
    ) 
    return prefactor * sc.lpmv(m, l, kz/k+0j, *args, **kwargs)

 
_periodic_to_scw = np.vectorize(_periodic_to_scw)


def periodic_to_scw(kz, m, l, mu, k, area, *args, **kwargs):
    """periodic_to_scw(kz, m, l, mu, k, area)
    
    Convert periodic spherical waves into cylindrical waves.

    Return the coefficient for the basis change from spherical waves on a lattice
    to cylindrical waves. For multiple positions, only the diagonal values (corresponding
    to identical positions) are returned. 

    Args:
        kz (float or array_like): Z-component of the wave vector of cylinbdrical waves.
        m (int or array_like): Order or cylinbdrical waves.
        l (int or array_like): Degree of spherical waves.
        mu (int or array_like): Order of spherical waves.
        k (float or complex or array_like): Angular wavenumber.
        area (float or array_like): Unit cell area.

    Returns:
        complex
    """
    return _periodic_to_scw(kz, m, l, mu, k, area, *args, **kwargs)


"""Scalar basis sets and core array functionalities."""

import abc
from collections import namedtuple

import numpy as np

import treams._operators as op
import acoustotreams._operatorsacoustics as opa
import treams.lattice as la
from treams import util
from treams._lattice import Lattice, WaveVector
from acoustotreams._materialacoustics import AcousticMaterial

class ScalarBasisSet(util.OrderedSet, metaclass=abc.ABCMeta):
    """Parent class of scalar basis sets.

    It is the parent class for all basis sets used. They are expected to be an ordered
    sequence of the modes included in an expansion. Basis sets are expected to
    be immutable.
    """

    _names = ()
    """Names of the relevant parameters"""

    def __repr__(self):
        """String representation.

        Automatically generated when the attribute ``_names`` is defined.

        Returns:
            str
        """
        string = ",\n    ".join(f"{name}={i}" for name, i in zip(self._names, self[()]))
        return f"{self.__class__.__name__}(\n    {string},\n)"

    @classmethod
    @abc.abstractmethod
    def default(cls, *args, **kwargs):
        """Construct a default basis from parameters.

        Construct a basis set in a default order by giving few parameters.
        """
        raise NotImplementedError


class ScalarSphericalWaveBasis(ScalarBasisSet):

    r"""Basis of scalar spherical waves.

    Functions of the spherical wave basis are defined by their degree ``l`` 
    and order ``m``. If the basis is defined with respect to a single 
    expansion center, it is referred to as "global". If it contains multiple 
    expansion centers, it is referred to as "local". In a local basis, an
    additional position index ``pidx`` is used to link the modes to one of the
    specified ``positions``.

    Spherical waves can be separated into regular and singular modes, which
    describe incident and scattered fields, respectively. Therefore, 
    the basis modes refer to one of the functions, :func:`~acoustotreams.special.ssw_rPsi` 
    or :func:`~acoustotreams.special.ssw_Psi`, respectively.

    Args:
        modes (array_like): A tuple containing a list for each of ``l`` and ``m``,
            or ``pidx``, ``l`` and ``m``.
        positions (array_like, optional): The positions of the expansion centers 
            for the specified modes (m). Defaults to ``[[0, 0, 0]]``.

    Attributes:
        pidx (array_like): Integer referring to a row in :attr:`positions`.
        l (array_like): Degree as an integer :math:`l \geq 0`
        m (array_like): Order as an integer :math:`m \leq |l|`.
    """

    _names = ("pidx", "l", "m")

    def __init__(self, modes, positions=None):
        """Initalization."""
        tmp = []
        if isinstance(modes, np.ndarray):
            modes = modes.tolist()
        for m in modes:
            if m not in tmp:
                tmp.append(m)
        modes = tmp
        if len(modes) == 0:
            pidx = []
            l = []  # noqa: E741
            m = []
        elif len(modes[0]) == 2:
            l, m = (*zip(*modes),)
            pidx = np.zeros_like(l)
        elif len(modes[0]) == 3:
            pidx, l, m = (*zip(*modes),)
        else:
            raise ValueError("invalid shape of modes")

        if positions is None:
            positions = np.zeros((1, 3))
        positions = np.array(positions, float)
        if positions.ndim == 1:
            positions = positions[None, :]
        if positions.ndim != 2 or positions.shape[1] != 3:
            raise ValueError(f"invalid shape of positions {positions.shape}")

        self.pidx, self.l, self.m = [
            np.array(i, int) for i in (pidx, l, m)
        ]
        for i, j in ((self.pidx, pidx), (self.l, l), (self.m, m)):
            i.flags.writeable = False
            if np.any(i != j):
                raise ValueError("parameters must be integer")
        if np.any(self.l < 0):
            raise ValueError("'l' must be a non-negative integer")
        if np.any(self.l < np.abs(self.m)):
            raise ValueError("'|m|' cannot be larger than 'l'")
        if np.any(self.pidx >= len(positions)):
            raise ValueError("undefined position is indexed")

        self._positions = positions
        self._positions.flags.writeable = False
        self.lattice = self.kpar = None 

    def __len__(self):
        """Number of modes."""
        return len(self.l)     

    @property
    def positions(self):
        """Positions of the modes' expansion centers.

        The positions are an immutable (N, 3)-array. Each row corresponds to a point in
        the three-dimensional Cartesian space.
        """
        return self._positions

    def __repr__(self):
        """String representation."""
        positions = "positions=" + str(self.positions).replace("\n", ",")
        return f"{super().__repr__()[:-1]}    {positions},\n)"

    @property
    def isglobal(self):
        """Basis is defined with respect to a single (global) expansion center.

        Returns:
            bool
        """
        return len(self) == 0 or np.all(self.pidx == self.pidx[0])

    def __getattr__(self, key):
        dct = {"l": "l", "m": "m", "p": "pidx"}
        try:
            return tuple(getattr(self, dct[k]) for k in key)
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'"
            ) from None

    def __getitem__(self, idx):
        """Get a subset of the basis.

        This function supports indexing the basis set using an integer, 
        a slice, an ellipsis, an empty tuple, or a sequence of integers 
        or booleans. Indexing with an integer or an empty tuple returns 
        a tuple; all other index types return a new basis set.

        Alternatively, the string "plm", or "lm" can be used to select 
        only a corresponding subset of the attrubutes :attr:`pidx`, :attr:`l`, 
        and :attr:`m`.
        """
        res = self.pidx[idx], self.l[idx], self.m[idx]
        if isinstance(idx, (int, np.integer)) or (
            isinstance(idx, tuple) and len(idx) == 0
        ):
            return res
        return type(self)(zip(*res), self.positions)

    def __eq__(self, other):
        """Compare basis sets.

        Two basis sets are considered equal if they contain the same modes 
        in the same order and share identical expansion centers :attr:`positions`.
        """
        try:
            return self is other or (
                np.array_equal(self.pidx, other.pidx)
                and np.array_equal(self.l, other.l)
                and np.array_equal(self.m, other.m)
                and np.array_equal(self.positions, other.positions)
            )
        except AttributeError:
            return False

    @classmethod
    def default(cls, lmax, nmax=1, positions=None):
        """Default basis for a given maximum multipolar degree.

        By default, modes are grouped into separate blocks for each position 
        index in ascending order. Within each block, modes are sorted by 
        degree :math:`l` (lowest first) and then by order :math:m from 
        :math:`m = -l` to :math:`m = l`.

        Example:
            >>> acoustotreams.ScalarSphericalWaveBasis.default(2)
            ScalarSphericalWaveBasis(
                pidx=[0 0 0 0 0 0 0 0 0],
                l=[0 1 1 1 2 2 2 2 2],
                m=[ 0 -1  0  1 -2 -1  0  1  2],
                positions=[[0. 0. 0.]],
            )
            >>> acoustotreams.ScalarSphericalWaveBasis.default(0, 2, [[0, 0, 1.], [0, 0, -1.]])
            ScalarSphericalWaveBasis(
                pidx=[0 1],
                l=[0 0],
                m=[0 0],
                positions=[[ 0.  0.  1.], [ 0.  0. -1.]],
            )

        Args:
            lmax (int): Maximum multipolar degree.
            nmax (int, optional): Number of positions. Defaults to 1.
            positions (array_like of float, optional): Positions of the expansion centers (m). Defaults to [[0, 0, 0]].
        """
        modes = [
            [n, l, m]
            for n in range(0, nmax)
            for l in range(0, lmax + 1)  # noqa: E741
            for m in range(-l, l + 1)
        ]
        return cls(modes, positions=positions)

    @staticmethod
    def defaultlmax(dim, nmax=1):
        """Calculate the maximum degree for a given number of modes.

        Return the estimated maximum value of `l` given the dimension of the T-matrix. 
        This is the inverse of :meth:defaultdim. A value of zero is returned for empty T-matrices.

        Example:
            >>> acoustotreams.ScalarSphericalWaveBasis.defaultlmax(len(acoustotreams.ScalarSphericalWaveBasis.default(3)))
            3

        Args:
            dim (int): Dimension of the T-matrix, respectively number of modes.
            nmax (int, optional): Number of scatterers. Defaults to 1.

        Returns:
            int
        """
        res = np.sqrt(dim / nmax) - 1
        res_int = int(np.rint(res))
        if np.abs(res - res_int) > 1e-8 * np.maximum(np.abs(res), np.abs(res_int)):
            raise ValueError("cannot estimate the default lmax")
        return res_int

    @staticmethod
    def defaultdim(lmax, nmax=1):
        """Calculate the number of modes for a given mulipolar degree.

        Return the dimesion of the T-matrix given the maximum value of `l`. 
        This is the inverse of :meth:`defaultlmax`. A value of lmax=-1 is allowed.

        Args:
            lmax (int): Maximum multipolar degree
            nmax (int, optional): Number of scatterers. Defaults to `1`

        Returns:
            int
        """
        # lmax=-1 is allowed and won't give an error
        if lmax < -1 or nmax < 0:
            raise ValueError("maximal degree must be non-negative")
        return (lmax + 1) * (lmax + 1) * nmax

    @classmethod
    def _from_iterable(cls, it, positions=None):
        if isinstance(cls, ScalarSphericalWaveBasis):
            positions = cls.positions if positions is None else positions
            cls = type(cls)
        obj = cls(it, positions=positions)
        return obj    
    

class ScalarCylindricalWaveBasis(ScalarBasisSet):
    r"""Basis of scalar cylindrical waves.

    Functions of the cylindrical wave basis are defined by the Z-components of the wave
    vector ``kz`` and the order  ``m``. If the basis is defined with respect to a single 
    expansion center, it is referred to as "global". If it contains multiple expansion
    centers, it is referred to as "local". In a local basis, an additional position index 
    ``pidx`` is used to link the modes to one of the specified ``positions``.

    Cylindrical waves can be separated into regular and singular modes, which describe 
    incident and scattered fields, respectively. Thus, the basis modes refer to one of 
    the functions :func:`~acoustotreams.special.scw_rPsi` or :func:`~acoustotreams.special.scw_Psi`,
    respectively.

    Args:
        modes (array_like): A tuple containing a list for each of ``kz`` and ``m``,
            or ``pidx``, ``kz`` and ``m``.
        positions (array_like, optional): Positions of the expansion centers for the specified
            modes (m). Defaults to ``[[0, 0, 0]]``.

    Attributes:
        pidx (array_like): Integer referring to a row in :attr:`positions`.
        kz (array_like): Real-valued Z-component of the wave vector (rad/m).
        m (array_like): Integer order.
    """

    _names = ("pidx", "kz", "m")

    def __init__(self, modes, positions=None):
        """Initalization."""
        tmp = []
        if isinstance(modes, np.ndarray):
            modes = modes.tolist()
        for m in modes:
            if m not in tmp:
                tmp.append(m)
        modes = tmp
        if len(modes) == 0:
            pidx = []
            kz = []
            m = []
        elif len(modes[0]) == 2:
            kz, m = (*zip(*modes),)
            pidx = np.zeros_like(m)
        elif len(modes[0]) == 3:
            pidx, kz, m = (*zip(*modes),)
        else:
            raise ValueError("invalid shape of modes")

        if positions is None:
            positions = np.zeros((1, 3))
        positions = np.array(positions, float)
        if positions.ndim == 1:
            positions = positions[None, :]
        if positions.ndim != 2 or positions.shape[1] != 3:
            raise ValueError(f"invalid shape of positions {positions.shape}")

        self.pidx, self.m = [np.array(i, int) for i in (pidx, m)]
        self.kz = np.array(kz, float)
        self.kz.flags.writeable = False
        for i, j in ((self.pidx, pidx), (self.m, m)):
            i.flags.writeable = False
            if np.any(i != j):
                raise ValueError("parameters must be integer")
        if np.any(self.pidx >= len(positions)):
            raise ValueError("undefined position is indexed")

        self._positions = positions
        self._positions.flags.writeable = False

        self.lattice = self.kpar = None
        if len(self.kz) > 0 and np.all(self.kz == self.kz[0]):
            self.kpar = WaveVector(self.kz[0])
    
    def __len__(self):
        """Number of modes."""
        return len(self.m) 

    @property
    def positions(self):
        """Positions of the modes' expansion centers.

        The positions are an immutable (N, 3)-array. Each row corresponds to a point in
        the three-dimensional Cartesian space.
        """
        return self._positions

    def __repr__(self):
        """String representation."""
        positions = "positions=" + str(self.positions).replace("\n", ",")
        return f"{super().__repr__()[:-1]}    {positions},\n)"

    @property
    def isglobal(self):
        """Basis is defined with respect to a single (global) expansion center.

        Returns:
            bool
        """
        return len(self) == 0 or np.all(self.pidx == self.pidx[0])

    def __getattr__(self, key):
        dct = {"z": "kz", "m": "m", "p": "pidx"}
        try:
            return tuple(getattr(self, dct[k]) for k in key)
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'"
            ) from None

    def __getitem__(self, idx):
        """Get a subset of the basis.

        This function supports indexing the basis set using an integer, a slice, 
        an ellipsis, an empty tuple, or a sequence of integers or booleans. Indexing 
        with an integer or an empty tuple returns a tuple; all other index types 
        return a new basis set.

        Alternatively, the string "pkzm" or "kzm" can be used to select a 
        corresponding subset of the attributes :attr:`pidx`, :attr:`kz`, and :attr:`m`.
        """
        res = self.pidx[idx], self.kz[idx], self.m[idx]
        if isinstance(idx, (int, np.integer)) or (isinstance(idx, tuple) and idx == ()):
            return res
        return type(self)(zip(*res), self.positions)

    def __eq__(self, other):
        """Compare basis sets.

        Two basis sets are considered equal if they contain the same modes in the same 
        order and share identical expansion centers :attr:`positions`.
        """
        try:
            return self is other or (
                np.array_equal(self.pidx, other.pidx)
                and np.array_equal(self.kz, other.kz)
                and np.array_equal(self.m, other.m)
                and np.array_equal(self.positions, other.positions)
            )
        except AttributeError:
            return False

    @classmethod
    def default(cls, kzs, mmax, nmax=1, positions=None):
        """Default basis for given Z-components of the wave vector and a given maximum order.

        By default, modes are grouped into separate blocks for each position 
        index in ascending order. Within each block, modes are sorted by 
        the Z-component of the wave vector :math:`k_z` and then by order :math:m 
        in asceding order.

        Example:
            >>> acoustotreams.ScalarCylindricalWaveBasis.default([-0.5, 0.5], 1)
            ScalarCylindricalWaveBasis(
                pidx=[0 0 0 0 0 0],
                kz=[-0.5 -0.5 -0.5 0.5  0.5  0.5],
                m=[-1  0  1 -1  0  1],
                positions=[[0. 0. 0.]],
            )
            >>> acoustotreams.ScalarCylindricalWaveBasis.default([0], 1, 2, [[1., 0, 0], [-1., 0, 0]])
            ScalarCylindricalWaveBasis(
                pidx=[0 0 0 1 1 1],
                kz=[0. 0. 0. 0. 0. 0.],
                m=[-1  0  1 -1  0  1],
                positions=[[ 1.  0.  0.], [-1.  0.  0.]],
            )

        Args:
            kzs (array_like of float): Z-components of the wave vector (rad/m).
            mmax (int): Maximum value of order.
            nmax (int, optional): Number of positions. Defaults to 1.
            positions (array_like of float, optional): Positions of the expansion centers (m). Defaults to [[0, 0, 0]].
        """
        kzs = np.atleast_1d(kzs)
        if kzs.ndim > 1:
            raise ValueError(f"kzs has dimension larger than one: '{kzs.ndim}'")
        modes = [
            [n, kz, m]
            for n in range(nmax)
            for kz in kzs
            for m in range(-mmax, mmax + 1)
        ]
        return cls(modes, positions=positions)

    @classmethod
    def diffr_orders(cls, kz, mmax, lattice, bmax, nmax=1, positions=None):
        """Creates a basis set for a system periodic in the z-direction.

        Example:
           >>> acoustotreams.ScalarCylindricalWaveBasis.diffr_orders(0.1, 1, 2 * np.pi, 1)
           ScalarCylindricalWaveBasis(
               pidx=[0 0 0 0 0 0 0 0 0],
               kz=[-0.9 -0.9 -0.9  0.1  0.1  0.1  1.1  1.1  1.1],
               m=[-1  0  1 -1  0  1 -1  0  1],
               positions=[[0. 0. 0.]],
           )

        Args:
            kz (float): Z-component of the wave vector. Ideally, it is defined within the 
                first Brillouin zone (use :func:`treams.misc.firstbrillouin1d`).
            mmax (int): Maximum value of the order.
            lattice (:class:`acoustotreams.Lattice` or float): Lattice definition or pitch.
            bmax (float): Maximum length of reciprocal-lattice vectors that defines 
                a maximum momentum transfer from the given value of `kz`.
            nmax (int, optional): Number of expansion centers.
            positions (array_like of float, optional): Positions of the expansion centers (m). Defaults to [[0, 0, 0]].
        """
        lattice = Lattice(lattice)
        lattice_z = Lattice(lattice, "z")
        nkz = np.floor(np.abs(bmax / lattice_z.reciprocal))
        kzs = kz + np.arange(-nkz, nkz + 1) * lattice_z.reciprocal
        res = cls.default(kzs, mmax, nmax, positions=positions)
        res.lattice = lattice
        res.kpar = WaveVector(kz)
        return res

    @classmethod
    def _from_iterable(cls, it, positions=None):
        if isinstance(cls, ScalarCylindricalWaveBasis):
            positions = cls.positions if positions is None else positions
            lattice = cls.lattice
            kpar = cls.kpar
            cls = type(cls)
        else:
            lattice = kpar = None
        obj = cls(it, positions=positions)
        obj.lattice = lattice
        obj.kpar = kpar
        return obj

    @staticmethod
    def defaultmmax(dim, nkz=1, nmax=1):
        """Calculate the maximum order for a given number of modes.

        Return the estimated maximum value of `m` given the dimension of the T-matrix.
        This is the inverse of :meth:`defaultdim`. A value of zero is returned for 
        empty T-matrices.

        Example:
            >>> acoustotreams.ScalarCylindricalWaveBasis.defaultmmax(len(acoustotreams.ScalarCylindricalWaveBasis.default([0], 2)), 1)
            2

        Args:
            dim (int): Dimension of the T-matrix, respectively number of modes.
            nkz (int, optional): Number of Z-components of the wave vector. Defaults to 1.
            nmax (int, optional): Number of scatterers. Defaults to 1.

        Returns:
            int
        """
        if dim % (nkz * nmax) != 0:
            raise ValueError("cannot estimate the default mmax")
        dim = dim // (nkz * nmax)
        if (dim - 1) % 2 != 0:
            raise ValueError("cannot estimate the default mmax")
        return (dim - 1) // 2

    @staticmethod
    def defaultdim(nkz, mmax, nmax=1):
        """Calculate the number of modes for a given mulipolar order.

        Return the dimension of the T-matrix given the maximum value of 
        `m` and number of `k_z` components. A value of zero is allowed.

        Args:
            nkz (int): Number of Z-components of the wave vector.
            mmax (int): Maximum value of the order.
            nmax (int, optional): Number of scatterers. Defaults to 1.

        Returns:
            int
        """
        if nkz < 0 or mmax < 0:
            raise ValueError("maximal order must be positive")
        return (2 * mmax + 1) * nkz * nmax
    
    
class ScalarPlaneWaveBasis(ScalarBasisSet):
    """Scalar plane wave basis parent class."""

    isglobal = True


class ScalarPlaneWaveBasisByUnitVector(ScalarPlaneWaveBasis):
    """Scalar plane wave basis.

    A plane-wave basis is defined by a collection of wave vectors specified by their
    Cartesian components ``qx``, ``qy``, and ``qz`` normalized to
    :math:`q_x^2 + q_y^2 + q_z^2 = 1`.

    Plane waves can refer to:func:`~acoustotreams.special.spw_Psi`.

    Args:
        modes (array_like): A tuple containing a list for each of ``qx``, ``qy``,
            and ``qz``.

    Attributes:
        qx (array_like): X-component of the normalized wave vector.
        qy (array_like): Y-component of the normalized wave vector.
        qz (array_like): Z-component of the normalized wave vector.
    """

    _names = ("qx", "qy", "qz")
    """A scalar plane-wave basis is always global."""

    def __init__(self, modes):
        """Initialization."""
        tmp = []
        if isinstance(modes, np.ndarray):
            modes = modes.tolist()
        for m in modes:
            if m not in tmp:
                tmp.append(m)
        modes = tmp
        if len(modes) == 0:
            qx = []
            qy = []
            qz = []
        elif len(modes[0]) == 3:
            qx, qy, qz = (*zip(*modes),)
        else:
            raise ValueError("invalid shape of modes")

        qx, qy, qz = (np.array(i) for i in (qx, qy, qz))
        norm = np.emath.sqrt(qx * qx + qy * qy + qz * qz)
        norm[np.abs(norm - 1) < 1e-14] = 1
        qx, qy, qz = (np.true_divide(i, norm) for i in (qx, qy, qz))
        for i in (qx, qy, qz):
            i.flags.writeable = False
            if i.ndim > 1:
                raise ValueError("invalid shape of parameters")
        self.qx, self.qy, self.qz = qx, qy, qz
        self.lattice = self.kpar = None

    def __len__(self):
        """Number of modes."""
        return len(self.qx) 

    def __getattr__(self, key):
        dct = {"x": "qx", "y": "qy", "z": "qz"}
        try:
            return tuple(getattr(self, dct[k]) for k in key)
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'"
            ) from None

    def __getitem__(self, idx):
        """Get a subset of the basis.

        This function supports indexing the basis set using an integer, a slice, an ellipsis, 
        an empty tuple, or a sequence of integers or booleans. Indexing with an integer or 
        an empty tuple returns a tuple; all other index types return a new basis set.

        Alternatively, the string "xyz", "xy" or "z" can be used to select only a
        corresponding subset of the attributes :attr:`qx`, :attr:`qy`, and :attr:`qz`.
        """
        res = self.qx[idx], self.qy[idx], self.qz[idx]
        if isinstance(idx, (int, np.integer)) or (isinstance(idx, tuple) and idx == ()):
            return res
        return type(self)(zip(*res))

    @classmethod
    def default(cls, kvecs):
        """Default basis from the given wave vectors.

        Example:
            >>> acoustotreams.ScalarPlaneWaveBasisByUnitVector.default([[0, 0, 5], [0, 3, 4]])
            ScalarPlaneWaveBasisByUnitVector(
                qx=[0. 0.],
                qy=[0.  0.6],
                qz=[1.  0.8],
            )

        Args:
            kvecs (array_like): Wave vectors in Cartesian coordinates (rad/m).
        """
        kvecs = np.atleast_2d(kvecs)
        modes = np.empty((kvecs.shape[0], 3), kvecs.dtype)                           
        modes = kvecs
        return cls(modes)

    @classmethod
    def _from_iterable(cls, it):
        if isinstance(cls, ScalarPlaneWaveBasisByUnitVector):
            lattice = cls.lattice
            kpar = cls.kpar
            cls = type(cls)
        else:
            lattice = kpar = None
        obj = cls(it)
        obj.lattice = lattice
        obj.kpar = kpar
        return obj

    def __eq__(self, other):
        """Compares basis sets.

        Two basis sets are considered equal if they contain the same modes in the same order.
        """
        try:
            return self is other or (
                np.array_equal(self.qx, other.qx)
                and np.array_equal(self.qy, other.qy)
                and np.array_equal(self.qz, other.qz)
            )
        except AttributeError:
            return False

    def bycomp(self, k0, alignment="xy", material=AcousticMaterial()):
        """Creates an instance of :class:`ScalarPlaneWaveBasisByComp`.

        The plane-wave basis is changed to a partial basis, where only two (real-valued)
        wave-vector components are specified. A third component is then inferred from the
        dispersion relation, which depends on the wavenumber, material and
        ``modetype``, which determines the sign of the third component.

        Args:
            alignment (str, optional): Wave-vector components included in the
                partial basis. Defaults to "xy", other permitted values are "yz" and
                "zx".
            k0 (float, optional): Angular wavenumber in the air (rad/m). 
                Verifies that the current basis satisfies the dispersion relation.
            material (:class:`acoustotreams.AcousticMaterial` or tuple): Material definition. 
                Defaults to the air.
        """
        ks = material.ks(k0)
        if alignment in ("xy", "yz", "zx"):
            kpars = [ks * getattr(self, "q" + s) for s in alignment]
        else:
            raise ValueError(f"invalid alignment '{alignment}'")
        obj = ScalarPlaneWaveBasisByComp(zip(*kpars), alignment)
        obj.lattice = self.lattice
        obj.kpar = self.kpar
        return obj

    def kvecs(self, k0, material=AcousticMaterial(), modetype=None):
        """Wave vectors.

        Args:
            k0 (float): Angular wavenumber in the air (rad/m).
            material (:class:`acoustotreams.AcousticMaterial` or tuple, optional): Material 
                definition. Defaults to the air.
            modetype (optional): Currently unused for this class.
        """
        # TODO: check kz depending on modetype (alignment?)
        ks = AcousticMaterial(material).ks(k0)
        return ks * self.qx, ks * self.qy, ks * self.qz

    def permute(self, n=1):
        n = n % 3
        lattice = None if self.lattice is None else self.lattice.permute(n)
        kpar = None if self.kpar is None else self.kpar.permute(n)
        qx, qy, qz = self.qx, self.qy, self.qz
        for _ in range(n):
            qx, qy, qz = qz, qx, qy
        obj = type(self)(zip(*(qx, qy, qz)))
        obj.lattice = lattice
        obj.kpar = kpar
        return obj
    

class ScalarPlaneWaveBasisByComp(ScalarPlaneWaveBasis):
    """Partial scalar plane wave basis.

    A partial plane-wave basis is defined by two wave-vector components out of all three
    Cartesian wave-vector components ``kx``, ``ky``, and ``kz``. 
    Which two components are given is specified in the :attr:`alignment`. This
    basis is mostly used for stratified media that are periodic or uniform in the two
    alignment directions, such that the given wave-vector components correspond to the
    diffraction orders.

    Scalar plane waves refer to :func:`~acoustotreams.special.spw_Psi`.

    Args:
        modes (array_like): A tuple containing a list for each of ``k1`` and ``k2``.
        alignment (str, optional): Wave-vector components included in the
                partial basis. Defaults to "xy", other possible values are "yz" and "zx".

    Attributes:
        alignment (str): Alignment of the partial basis.
    """

    def __init__(self, modes, alignment="xy"):
        """Initialization."""
        if isinstance(modes, np.ndarray):
            modes = modes.tolist()
        tmp = []
        for m in modes:
            if m not in tmp:
                tmp.append(m)
        modes = tmp
        if len(modes) == 0:
            kx = []
            ky = []
        elif len(modes[0]) == 2:
            kx, ky = (*zip(*modes),)
        else:
            raise ValueError("invalid shape of modes")

        self._kx, self._ky = [np.real(i) for i in (kx, ky)]
        for i, j in [(self._kx, kx), (self._ky, ky)]:
            i.flags.writeable = False
            if i.ndim > 1:
                raise ValueError("invalid shape of parameters")
            if np.any(i != j):
                raise ValueError("invalid value for parameter, must be real")

        if alignment in ("xy", "yz", "zx"):
            self._names = (*("k" + i for i in alignment),)
        else:
            raise ValueError(f"invalid alignment '{alignment}'")

        self.alignment = alignment
        self.lattice = self.kpar = None

    def __len__(self):
        """Number of modes."""
        return len(self.kx)

    def permute(self, n=1):
        n = n % 3
        lattice = None if self.lattice is None else self.lattice.permute(n)
        kpar = None if self.kpar is None else self.kpar.permute(n)
        alignments = {"xy": "yz", "yz": "zx", "zx": "xy"}
        alignment = self.alignment
        for _ in range(n):
            alignment = alignments[alignment]
        obj = self._from_iterable(self, alignment=alignment)
        obj.lattice = lattice
        obj.kpar = kpar
        return obj

    @property
    def kx(self):
        """X-components of the wave vector.

        If the components are not specified, `None` is returned.
        """
        if self.alignment == "xy":
            return self._kx
        if self.alignment == "zx":
            return self._ky
        return None

    @property
    def ky(self):
        """Y-components of the wave vector.

        If the components are not specified, `None` is returned.
        """
        if self.alignment == "xy":
            return self._ky
        if self.alignment == "yz":
            return self._kx
        return None

    @property
    def kz(self):
        """Z-components of the wave vector.

        If the components are not specified, `None` is returned.
        """
        if self.alignment == "yz":
            return self._ky
        if self.alignment == "zx":
            return self._kx
        return None

    def __getattr__(self, key):
        dct = {"x": "kx", "y": "ky", "z": "kz"}
        try:
            return tuple(getattr(self, dct[k]) for k in key)
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'"
            ) from None

    def __getitem__(self, idx):
        """Get a subset of the basis.

        This function supports indexing the basis set using an integer, a slice, 
        an ellipsis, an empty tuple, or a sequence of integers or booleans. 
        Indexing with an integer or an empty tuple returns a tuple; 
        all other index types return a new basis set.
        """
        res = self._kx[idx], self._ky[idx]
        if isinstance(idx, (int, np.integer)) or (isinstance(idx, tuple) and idx == ()):
            return res
        return type(self)(zip(*res))

    @classmethod
    def default(cls, kpars, alignment="xy"):
        """Default basis from given wave vectors.

        Example:
            >>> acoustotreams.ScalarPlaneWaveBasisByComp.default([[0, 0], [0, 3]])
            ScalarPlaneWaveBasisByComp(
                kx=[0 0],
                ky=[0 3],
            )

        Args:
            kpars (array_like): Two Cartesian components of the wave vectors (rad/m).
            alignment (str, optional): Wave-vector components included in the
                partial basis. Defaults to "xy", other possible values are "yz" and "zx".

        """
        kpars = np.atleast_2d(kpars)
        modes = kpars
        return cls(modes, alignment=alignment)

    @classmethod
    def diffr_orders(cls, kpar, lattice, bmax):
        """Creates a basis set for a two-dimensional periodic system.

        The reciprocal lattice of the given lattice is used to determine 
        all diffraction orders within the defined maximum radius of the circle 
        in the reciprocal space.

        Example:
            >>> acoustotreams.ScalarPlaneWaveBasisByComp.diffr_orders([0, 0], acoustotreams.Lattice.square(2 * np.pi), 1)
            ScalarPlaneWaveBasisByComp(
                kx=[ 0.  0.  0.  1. -1.],
                ky=[ 0.  1. -1.  0.  0.],
            )

        Args:
            kpar (float): In-plane wave-vector components (rad/m). Ideally, they are defined within the
                first Brillouin zone (use :func:`misc.firstbrillouin2d`).
            lattice (:class:`acoustotreams.Lattice` or float): Lattice definition or pitch.
            bmax (float): Maximum length of reciprocal-lattice vectors that defines a maximum
                momentum transfer with respect to `kpar`.
        """
        lattice = Lattice(lattice)
        if lattice.dim != 2:
            raise ValueError("invalid lattice dimensions")
        latrec = lattice.reciprocal
        kpars = kpar + la.diffr_orders_circle(latrec, bmax) @ latrec
        obj = cls.default(kpars, alignment=lattice.alignment)
        obj.lattice = lattice
        obj.kpar = WaveVector(kpar, alignment=lattice.alignment)
        return obj

    @classmethod
    def _from_iterable(cls, it, alignment="xy"):
        if isinstance(cls, ScalarPlaneWaveBasisByComp):
            alignment = cls.alignment if alignment is None else alignment
            lattice = cls.lattice
            kpar = cls.kpar
            cls = type(cls)
        else:
            lattice = kpar = None
        obj = cls(it, alignment)
        obj.lattice = lattice
        obj.kpar = kpar
        return obj

    def __eq__(self, other):
        """Compares basis sets.

        Two basis sets are considered equal if they contain the same modes in the same order
        """
        try:
            skx, sky, skz = self.kx, self.ky, self.kz
            okx, oky, okz = other.kx, other.ky, other.kz
            return self is other or (
                (np.array_equal(skx, okx) or (skx is None and okx is None))
                and (np.array_equal(sky, oky) or (sky is None and oky is None))
                and (np.array_equal(skz, okz) or (skz is None and okz is None))
            )
        except AttributeError:
            return False

    def byunitvector(self, k0, material=AcousticMaterial(), modetype="up"):
        """Creates an instance of complete basis :class:`ScalarPlaneWaveBasis`.

        A plane-wave basis is considered complete when all three Cartesian components
        are defined for each mode. A third Cartesian component is calculated from
        the specified wavenumber, material, and modetype. The modetype "up" 
        (resp. "down") corresponds to waves propagating in the positive (resp. negative)
        direction with respect to the Cartesian axis that is orthogonal to the
        alignment plane.

        Args:
            k0 (float): Angular wavenumber in the air (rad/m).
            material (:class:`~acoustotreams.AcousticMaterial` or tuple, optional): Material 
                definition. Defaults to the air.
            modetype (str, optional): Propagation direction. Defaults to "up".
        """
        if modetype not in ("up", "down"):
            raise ValueError("modetype not recognized")
        material = AcousticMaterial(material)
        kx = self._kx
        ky = self._ky
        kz = material.kzs(k0, kx, ky) * (2 * (modetype == "up") - 1)
        if self.alignment == "yz":
            kx, ky, kz = kz, kx, ky
        elif self.alignment == "zy":
            kx, ky, kz = ky, kz, kx
        obj = ScalarPlaneWaveBasisByUnitVector(zip(kx, ky, kz))
        obj.lattice = self.lattice
        obj.kpar = self.kpar
        return obj

    def kvecs(self, k0, material=AcousticMaterial(), modetype="up"):
        """Wave vectors.

        Args:
            k0 (float): Angular wavenumber in the air (rad/m).
            material (:class:`~acoustotreams.AcousticMaterial` or tuple, optional): Material
                definition. Defaults to the air.
            modetype (str, optional): Propagation direction. Defaults to "up".
        """
        if modetype not in ("up", "down"):
            raise ValueError("modetype not recognized")
        material = AcousticMaterial(material)
        kx = self._kx
        ky = self._ky
        kz = material.kzs(k0, kx, ky) * (2 * (modetype == "up") - 1)
        if self.alignment == "yz":
            return kz, kx, ky
        if self.alignment == "zx":
            return ky, kz, kx
        return kx, ky, kz


def _raise_basis_error(*args):
    raise TypeError("'basis' must be ScalarBasisSet")


class ScalarPhysicsDict(util.AnnotationDict):
    """Physics dictionary (for scalar waves).

    Derives from :class:`treams.util.AnnotationDict`. This dictionary has additionally
    several properties defined.

    Attributes:
        basis (:class:`ScalarBasisSet`): Basis set.
        k0 (float): Angular wavenumber in the air (rad/m).
        kpar (list): In-plane wave-vector components (rad/m). Usually, this is a list of length
            3 with its items corresponding to the Cartesian coordinates. Unspecified items
            aree set to `nan`.
        lattice (:class:`~acoustotreams.Lattice`): Lattice definition.
        material (:class:`~acoustotreams.AcousticMaterial`): Material definition.
        modetype (str): Mode type. For spherical and cylindrical waves, it can be
            "incident" and "scattered"; for partial plane waves, "up" or "down".
    """

    properties = {
        "basis": (
            ":class:`ScalarBasisSet`.",
            lambda x: isinstance(x, ScalarBasisSet),
            _raise_basis_error,
        ),
        "k0": ("Angular wavenumber.", lambda x: isinstance(x, float), float),
        "kpar": (
            "Wave-vector components tangential to the lattice.",
            lambda x: isinstance(x, WaveVector),
            WaveVector,
        ),
        "lattice": (
            ":class:`~acoustotreams.Lattice`.",
            lambda x: isinstance(x, Lattice),
            Lattice,
        ),
        "material": (
            ":class:`AcousticMaterial`.",
            lambda x: isinstance(x, AcousticMaterial),
            AcousticMaterial,
        ),
        "modetype": ("Mode type.", lambda x: isinstance(x, str), str),
    }
    """Special properties tracked by the ScalarPhysicsDict."""

    def __setitem__(self, key, val):
        """Sets item specified by key to the defined value.

        When overwriting an existing key an :class:`AnnotationWarning` is emitted.
        Avoid the warning by explicitly deleting the key first. The special attributes
        are cast to their corresponding types.

        Args:
            key (hashable): Key
            val : Value

        Warns:
            AnnotationWarning
        """
        if key not in self.properties:
            raise AttributeError(f"invalid key '{key}'")
        _, testfunc, castfunc = self.properties[key]
        if not testfunc(val):
            val = castfunc(val)
        super().__setitem__(key, val)


class AcousticsArray(util.AnnotatedArray):
    """Acoustics-aware array.

    An acoustics-aware array is a special type of :class`~treams.util.AnnotatedArray`.
    Additionally to keeping track of the annotations, it is enhanced by the ability to
    create suiting linear operators to perform tasks like rotations, translations, or
    expansions into different basis sets and by applying special rules for the
    physical properties of :class:`ScalarPhysicsDict` upon matrix multiplications, see also
    :meth:`__matmul__`.
    """

    _scales = {"basis"}

    pfield = op.OperatorAttribute(opa.PField)
    """Pressure field evaluation matrix, see also :class:`acoustotreams.PField`."""
    vfield = op.OperatorAttribute(opa.VField)
    """Velocity field evaluation matrix, see also :class:`acoustotreams.VField`."""
    pamplitudeff = op.OperatorAttribute(opa.PAmplitudeFF)
    """Far-field amplitude of pressure field evaluation matrix, see also :class:`acoustotreams.PAmplitudeFF`."""
    expand = op.OperatorAttribute(opa.Expand)
    """Expansion matrix, see also :class:`acoustotreams.Expand`."""
    expandlattice = op.OperatorAttribute(opa.ExpandLattice)
    """Lattice expansion matrix, see also :class:`acoustotreams.ExpandLattice`."""
    permute = op.OperatorAttribute(opa.Permute)
    """Permutation matrix, see also :class:`acoustotreams.Permute`."""
    rotate = op.OperatorAttribute(opa.Rotate)
    """Rotation matrix, see also :class:`acoustotreams.Rotate`."""
    translate = op.OperatorAttribute(opa.Translate)
    """Translation matrix, see also :class:`acoustotreams.Translate`."""

    def __init__(self, arr, ann=(), /, **kwargs):
        """Initialization."""
        super().__init__(arr, ann, **kwargs)
        self._check()

    @property
    def ann(self):
        """Array annotations."""
        # necessary to define the setter below
        return super().ann

    def __getattr__(self, key):
        try:
            return super().__getattr__(key)
        except AttributeError as err:
            if key in ScalarPhysicsDict.properties:
                return None
            raise err from None

    def __setattr__(self, key, val):
        if key in ScalarPhysicsDict.properties:
            val = (val,) * self.ndim if not isinstance(val, tuple) else val
            self.ann.as_dict[key] = val
        else:
            super().__setattr__(key, val)

    @ann.setter
    def ann(self, ann):
        """Sets array annotations.

        See also :meth:`treams.util.AnnotatedArray.__setitem__`.
        """
        self._ann = util.AnnotationSequence(*(({},) * self.ndim), mapping=ScalarPhysicsDict)
        self._ann.update(ann)

    def __repr__(self):
        """String representation.

        For a more managable output only the special physics properties are shown
        alongside the array itself.
        """
        repr_arr = "    " + repr(self._array)[5:-1].replace("\n  ", "\n") + ","        
        for key in ScalarPhysicsDict.properties:
            if getattr(self, key) is not None:
                repr_arr += f"\n    {key}={repr(getattr(self, key))},"
        return f"{self.__class__.__name__}(\n{repr_arr}\n)"
    
    def _check(self):
        """Run checks to validate the physical properties.

        The checks are run on the last two dimensions. They include:

            * Dispersion relation checks for :class:`ScalarPlaneWaveBasis` if `basis`, `k0`
              and `material` are defined`
            * All lattices explicitly given and in the basis hints must be compatible.
            * All tangential wave vector compontents must be compatible.
        """
        for a in self.ann[-2:]:
            k0 = a.get("k0")
            material = a.get("material")
            modetype = a.get("modetype")
            basis = a.get("basis", namedtuple("_basis", "lattice kpar")(None, None))

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Implements ufunc API.

        In addition to keeping track of the annotations, the special properties of an
        AcousticsArray are also "transparent" in matrix multiplications.
        """
        res = super().__array_ufunc__(ufunc, method, *inputs, **kwargs)
        if (
            ufunc is np.matmul
            and method == "__call__"
            and not isinstance(res, np.generic)
        ):
            axes = kwargs.get(
                "axes", [tuple(range(-min(np.ndim(i), 2), 0)) for i in inputs + (res,)]   #!!!!!!!!!!!
            )
            anns = [
                i.ann[ax] if hasattr(i, "ann") else [{}, {}]
                for i, ax in zip(inputs, axes)
            ]
            for name in ScalarPhysicsDict.properties:
                if name in anns[0][-1] and all(name not in a for a in anns[1]):
                    res.ann[axes[-1][-1]].setdefault(name, anns[0][-1][name])
                if name in anns[1][0] and all(name not in a for a in anns[0]):
                    res.ann[axes[-1][0]].setdefault(name, anns[1][0][name])
        return res


from setuptools import setup, find_packages



setup(

    name='acoustotreams',

    version='0.2.49',

    author='Nikita Ustimenko',

    author_email='nikita.ustimenko@kit.edu',

    description='A Python package for acoustic-wave scattering based on the T-matrix method',

    license = 'MIT',

    long_description=open('README.md').read(),

    long_description_content_type='text/markdown',

    url='https://github.com/NikUstimenko/acoustotreams',

    project_urls={
        "Bug Tracker": "https://github.com/NikUstimenko/acoustotreams/issues",
    },  

    packages=find_packages(),

    classifiers=[

        'Programming Language :: Python :: 3',

        'License :: OSI Approved :: MIT License',

        'Operating System :: OS Independent',

    ],

    python_requires='>=3.9,<3.14',

    install_requires=[
        "numpy<2.4",
        "scipy>=1.6,<1.14; python_version < '3.10'",
        "scipy>=1.14.1,<1.17; python_version >= '3.10'",
        "treams"
    ]

)



"""Scalar cylindrical wave module.

.. autosummary::
   :toctree:

   periodic_to_spw
   rotate
   to_ssw
   translate
   translate_periodic

"""

import acoustotreams.special as ats
import treams.special as sc
import numpy as np
from treams import lattice
import scipy.special as ss


def _translate_s(kz, mu, qz, m, krr, phi, z, *args, **kwargs):
    if abs(krr) < 1e-16 and abs(z) < 1e-16:
        return 0.0j
    return ats.tl_scw(kz, mu, qz, m, krr, phi, z, *args, **kwargs)

def _translate_r(kz, mu, qz, m, krr, phi, z, *args, **kwargs):
    return ats.tl_scw_r(kz, mu, qz, m, krr, phi, z, *args, **kwargs)


_translate_s = np.vectorize(_translate_s)
_translate_r = np.vectorize(_translate_r)


def translate(kz, mu, qz, m, krr, phi, z, singular=True, *args, **kwargs):
    """translate(kz, mu, qz, m, krr, phi, z, singular=True)
    
    Translation coefficients for scalar cylindrical modes.

    Returns the correct translation coefficients from :func:`acoustotreams.scw.tl_scw`
    and :func:`acoustotreams.scw.tl_scw_r` or combinations thereof for the specified modes and
    basis.

    Args:
        kz (float or array_like): Z-component of the wave vector of output modes.
        mu (int or array_like): Order of output modes.
        qz (float or array_like): Z component of the wave vector of input modes. Has units of kz.
        m (int or array_like): Order of input modes.
        krr (float or complex or array_like): Translation distance in units of 
            the radial component of the wave vector.
        phi (float or array_like): Azimuthal angle (rad).
        z (float or array_like): Z-coordinate. Has units of 1/kz.
        singular (bool, optional): If true, singular translation coefficients are used,
            else regular coefficients. Defaults to ``True``.

    Returns:
        complex or array_like
    """
    if singular:
        return _translate_s(kz, mu, qz, m, krr, phi, z, *args, **kwargs)
    return _translate_r(kz, mu, qz, m, krr, phi, z, *args, **kwargs)

def _rotate(kz, mu, qz, m, phi, *args, **kwargs):
    if (kz == qz) and (m == mu):
        return np.exp(-1j * m * phi, *args, **kwargs)
    else:
        return 0.+0.j
    
_rotate = np.vectorize(_rotate)

def rotate(kz, mu, qz, m, phi, *args, **kwargs):
    """rotate(kz, mu, qz, m, phi)

    Rotation coefficients for scalar cylindrical modes.

    Return the rotation coefficients or a combination thereof for the specified modes.

    Args:
        kz (float or array_like): Z-component of the wave vector of output modes.
        mu (int or array_like): Order of output modes.
        qz (float or array_like): Z-component of the wave vector of input modes. Has units of kz.
        m (int or array_like): Order of input modes.
        phi (float or array_like): Rotation angle (rad).

    Returns:
        complex or array_like
    """
    return _rotate(kz, mu, qz, m, phi, *args, **kwargs)

   
def _periodic_to_spw(kx, ky, kzpw, kzcw, m, a, *args, **kwargs):
   krho = np.sqrt(kx * kx + ky * ky)
   ky_s = ky
   if np.abs(kzcw - kzpw) < 1e-12:
      if ky_s == 0:
            ky_s = 1e-20 + 1e-20j
      elif np.imag(ky_s) < 0 or (np.imag(ky_s) == 0 and np.real(ky_s) < 0):
            ky_s = -ky_s
      if krho == 0:
         return 2 * np.power(-1j, m) / (np.abs(a) * ky_s)
      return 2 * np.power((-1j * kx + ky) / krho, m, *args, **kwargs) / (np.abs(a) * ky_s)
   else:
      return 0.0j + sc.lpmv(0, 1, 0, *args, **kwargs)

_periodic_to_spw = np.vectorize(_periodic_to_spw)

def periodic_to_spw(kx, ky, kzpw, kzcw, m, a, *args, **kwargs):
    """periodic_to_spw(kx, ky, kz, qz, m, a)

    Convert periodic scalar cylindrical waves into scalar plane waves.

    Return the coefficients for the basis change from a periodic arrangement of cylindrical
    waves to plane waves. For multiple positions, only the diagonal values (corresponding to
    identical positions) are returned.

    Args:
        kx (float or array_like): X-component of the wave vector of plane waves.
        ky (float or complex or array_like): Y-component of the wave vector of plane waves. Has units of kx.
        kz (float or array_like): Z-component of the wave vector of plane waves. Has units of kx.
        qz (float or array_like): Z-component of the waver vector of cylindrical waves. Has units of kx.
        m (int or array_like): Order of cylindrical waves.
        area (float or array_like): Unit cell area.

    Returns:
        complex or array_like

    """
    return _periodic_to_spw(kx, ky, kzpw, kzcw, m, a, *args, **kwargs)

def _to_ssw(l, m, kz, mu, k, *args, **kwargs):
    if m == mu:
        return (
            np.power(1j, l - m, *args, **kwargs)
            * np.sqrt(4 * np.pi * (2 * l + 1), *args, **kwargs)
            * np.sqrt(ss.gamma(l - m + 1) / ss.gamma(l + m + 1), *args, **kwargs)
            * sc.lpmv(m, l, kz/k+0j, *args, **kwargs)
            )
    return 0.0j + sc.lpmv(0, 1, 0, *args, **kwargs)

_to_ssw = np.vectorize(_to_ssw)

def to_ssw(l, m, kz, mu, k, *args, **kwargs):
    """to_ssw(l, m, kz, mu, k)
    
    Coefficients for the expansion of scalar cylindrical waves in scalar spherical waves.

    Return the coefficient for the basis change from cylindrical waves to 
    spherical waves. For multiple positions, only the diagonal values 
    (corresponding to identical positions) are returned.

    Args:
        l (int or array_like): Degree of spherical waves.
        m (int or array_like): Order of spherical waves.
        kz (float or array_like): Z-component of the wave vector of cylindrical waves.
        mu (int or array_like): Order of cylindrical waves.
        k (float or complex or array_like): Angular wavenumber in the medium. Has units of kz.
    
    Returns:
        complex or array_like
    """
    return _to_ssw(l, m, kz, mu, k, *args, **kwargs)


def translate_periodic(k, kpar, a, rs, out, in_=None, rsin=None, eta=0):
    """Translation coefficients for cylindrical waves on a lattice.

    Return the translation coefficents for the given modes on a lattice. The
    computations employ the fast converging Ewald summation from :mod:`treams.lattice`.

    Args:
        k (float or complex): Angular wavenumber in the medium.
        kpar ((D,)-array of float): Parallel component of the wave vector. Has units of k.
            Defines the dimension with `1 <= D <= 2`.
        a ((D,D)-array of float): Lattice vectors in each row of the array. Have units of 1/k.
        rs ((M, 3)-array of float): Shift vectors with respect to one lattice point. Have units of 1/k. 
        out (2- or 3-tuple of arrays): Output modes. If `len(out) == 2`, `out[0]` may be float 
            and `out[1]` must be integer. If `len(out) == 3`, `out[1]` may be float, while `out[0]` 
            and `out[2]` must be integer.
        in_ (2- or 3-tuple of arrays): Input modes. If None is given, they are equal to the output 
            modes. If `len(in_) == 2`, `in_[0]` may be float and `in_[1]` must be integer. 
            If `len(in_) == 3`, `in_[1]` may be float, while `in_[0]` and `in_[2]` must be integer.
        rsin (float): Shift vectors of the input modes. If None is given, they are equal
            to `rs`. Have units of 1/k.
        eta (float or complex, optional): Splitting parameter for the Ewald summation of
            lattice sums. If zero, an estimation for the optimal value is done. Defaults
            to zero.

    Returns:
        array_like of complex
    """
    if in_ is None:
        in_ = out
    out = (*(np.array(o) for o in out),)
    in_ = (*(np.array(i) for i in in_),)
    if len(out) < 2 or len(out) > 3:
        raise ValueError(f"invalid length of output modes {len(out)}, must be 2 or 3")
    elif len(out) == 2:
        out = (np.zeros_like(out[0]),) + out
    if len(in_) < 2 or len(in_) > 3:
        raise ValueError(f"invalid length of input modes {len(in_)}, must be 2 or 3")
    elif len(in_) == 2:
        in_ = (np.zeros_like(in_[0]),) + in_
    if rsin is None:
        rsin = rs
    modes = -out[2][:, None] + in_[2]
    k = np.array(k)
    k = k.reshape((-1,))
    krhos = np.sqrt((k[0] * k[0] - out[1][:, None] * in_[1]).astype(complex))  # todo: out not necessary this only simplifies krhos[compute] below
    krhos[krhos.imag < 0] = -krhos[krhos.imag < 0]
    compute = np.equal(out[1][:, None], in_[1])
    kpar = np.array(kpar)
    rs = np.array(rs)
    if rs.ndim == 1:
        rs = rs.reshape((1, -1))
    rsin = np.array(rsin)
    if rsin.ndim == 1:
        rsin = rsin.reshape((1, -1))
    rsdiff = -rs[out[0].astype(int), None, :] + rsin[in_[0].astype(int), :]
    if kpar.ndim == 0 or kpar.shape[-1] == 1:
        dlms = lattice.lsumcw1d_shift(modes[compute], krhos[compute], kpar, a, rsdiff[compute, :2], eta)
    else:
        dlms = lattice.lsumcw2d(modes[compute], krhos[compute], kpar, a, rsdiff[compute, :2], eta)
    res = np.zeros((out[0].shape[0], in_[0].shape[0]), complex)
    res[compute] = dlms
    res = res * np.exp(-1j * in_[1] * rsdiff[:, :, 2])
    return res
